import parsec as p
import re


def clean_host(x):
    x = re.sub(";.*", "", x.strip().lower())
    if "scrofa" in x:
        x = "swine"
    elif "pig" in x:
        x = "swine"
    elif "porcine" in x:
        x = "swine"
    elif "boar" in x:
        x = "swine"
    elif "sapiens" in x:
        x = "human"
    return x


p_host = p.regex(re.compile("swine|human", re.IGNORECASE))



import parsec as p
from octofludb.nomenclature import make_property, make_literal, define_subproperty
import rdflib
import sys
import re
import math
from octofludb.util import rmNone, log
from rdflib.namespace import XSD


class Token:
    parser = None
    group = None
    typename = "auto"
    class_predicate = None

    def __init__(self, text, field_name=None, na_str=[None]):
        self.matches = self.testOne(text, na_str=na_str)
        self.dirty = text
        self.field_name = field_name
        if self.matches is not None:
            self.clean = self.munge(self.matches)
        else:
            self.clean = None

    def munge(self, text):
        return text

    def choose_field_name(self):
        if self.field_name:
            return self.field_name
        else:
            return self.typename

    def as_uri(self):
        """
        Cast this token as a URI
        """
        return None  # default tokens cannot be URIs

    def as_predicate(self):
        return make_property(self.choose_field_name())

    def as_object(self):
        return self.as_literal()

    def object_of(self, g, uri):
        if uri and self.matches:
            g.add((uri, self.as_predicate(), self.as_object()))

    def as_literal(self):
        """
        Cast this token as a literal value
        """
        return rdflib.Literal(self.clean)

    def add_triples(self, g):
        """
        Add knowledge to the graph
        """
        return None
        # # FIXME: the following code unifies fields of the same type, but it can go too far
        #  if (
        #      self.field_name
        #      and self.field_name != self.typename
        #      and self.class_predicate
        #      and self.matches
        #  ):
        #      define_subproperty(self.as_predicate(), self.class_predicate, g)

    def relate(self, fields, g, levels=None):
        """
        Create links as desired between Tokens.
        """
        pass

    def __str__(self):
        return self.clean

    def __bool__(self):
        return self.matches is not None and self.matches != ""

    @classmethod
    def testOne(cls, item, na_str=[None]):
        """
        The item is a member of this type. In the base case anything
        that can be turned into a string is a member.
        """
        if item in na_str:
            return None
        try:
            return cls.parser.parse_strict(item)
        except p.ParseError:
            return None

    @classmethod
    def goodness(cls, items, na_str=[None]):
        matches = [
            (cls.testOne(item=x, na_str=na_str) != None)
            for x in items
            if x not in na_str
        ]
        if len(matches) > 0:
            return sum(matches) / len(matches)
        else:
            return 0


class Missing(Token):
    parser = lambda x: None
    typename = "missing"

    @classmethod
    def testOne(cls, item, na_str=[None]):
        return None


class Unknown(Token):
    typename = "unknown"
    parser = lambda x: x

    @classmethod
    def testOne(cls, item, na_str=[None]):
        try:
            if math.isnan(item):
                return None
        except TypeError:
            pass
        if item in na_str:
            return None
        else:
            return item


class Integer(Token):
    typename = "integer"
    parser = p.regex("[1-9]\\d*") ^ p.string("0")

    def as_literal(self):
        return rdflib.Literal(self.clean, datatype=XSD.integer)


class Double(Token):
    typename = "double"
    parser = (
        p.regex("0\\.\\d+")
        ^ p.regex("[1-9]\\d*\\.\\d+")
        ^ p.regex("[1-9]\\d*")
        ^ p.string("0")
    )

    def as_literal(self):
        return rdflib.Literal(self.clean, XSD.double)


class Boolean(Token):
    typename = "float"
    parser = p.regex("0|1|yes|no|true|false|y|n|t|f", flags=re.I)

    def munge(self, text):
        if text.lower() in ["1", "t", "true", "yes", "y"]:
            return "true"
        else:
            return "false"

    def as_literal(self):
        return rdflib.Literal(self.clean, XSD.boolean)


class Ignore(Token):
    typename = "ignore_me"
    parser = lambda x: None

    @classmethod
    def testOne(cls, item, na_str=[None]):
        return None


class Empty(Token):
    """If you want to throw out any field that is not recognized, make this the default"""

    typename = "empty"
    parser = p.regex(".*")

    def munge(self, text):
        return ""



from octofludb.classes import Datum, Phrase


def addTriples(g, triples):
    for triple in triples:
        g.add(triple)


def showTriple(xs):
    """
    This is mostly for diagnostics in the REPL and test
    """
    g = set()
    Phrase([Datum(x).data for x in xs]).connect(g)
    s = sorted([(str(s), str(p), str(o)) for s, p, o in g])
    return s



import parsec as p
import re
import rdflib
from typing import List
from collections import defaultdict


def wordset(words, label, f=lambda x: x.lower().replace(" ", "_")):
    """
    Create a log(n) parser for a set of n strings.
    @param "label" is a arbitrary name for the wordset that is used in error messages
    @param "f" is a function used to convert matching strings in the wordset and text (e.g to match on lower case).
    """
    d = defaultdict(set)
    # divide words into sets of words of the same length
    # this allows exact matches against the sets
    for word in words:
        d[len(word)].update([f(word)])
    # Convert this to a list of (<length>, <set>) tuples reverse sorted by
    # length. The parser must search for longer strings first to avoid matches
    # against prefixes.
    d = sorted(d.items(), key=lambda x: x[0], reverse=True)

    @p.Parser
    def wordsetParser(text, index=0):
        for k, v in d:
            if f(text[index : (index + k)]) in v:
                return p.Value.success(index + k, text[index : (index + k)])
        return p.Value.failure(
            index, f'a term "{f(text[index:(index+k)])}" not in wordset {d}'
        )

    return wordsetParser


def parse_match(parser, text):
    try:
        parser.parse_strict(text)
    except p.ParseError:
        return False
    return True


def regexWithin(regex, context: p.Parser):
    @p.Parser
    def regexWithinParser(text, index=0):
        try:
            contextStr = context.parse(text[index:])
        except p.ParseError:
            return p.Value.failure(index, "Context not found")
        m = re.search(regex, contextStr)
        if m:
            return p.Value.success(index + len(contextStr), m.group(0))
        else:
            return p.Value.failure(index, "Could not match regex")

    return regexWithinParser


def splitMatchFirst(psr: p.Parser, splitStr: str, text: str):
    """
    Rethink whether this is needed ...
    """
    fields = text.split(splitStr)
    for field in fields:
        try:
            return psr.parse_strict(field)
        except:
            continue
    return None



import re
from octofludb.spellcheck import make_flat_wordfinder

STATE_NAME2ABBR = {
    "alaska": "AK",
    "alabama": "AL",
    "arkansas": "AR",
    "arizona": "AZ",
    "california": "CA",
    "colorado": "CO",
    "connecticut": "CT",
    "district_of_columbia": "DC",
    "delaware": "DE",
    "florida": "FL",
    "georgia": "GA",
    "hawaii": "HI",
    "iowa": "IA",
    "idaho": "ID",
    "illinois": "IL",
    "indiana": "IN",
    "kansas": "KS",
    "kentucky": "KY",
    "louisiana": "LA",
    "massachusetts": "MA",
    "maryland": "MD",
    "maine": "ME",
    "michigan": "MI",
    "minnesota": "MN",
    "missouri": "MO",
    "mississippi": "MS",
    "montana": "MT",
    "north_carolina": "NC",
    "north_dakota": "ND",
    "nebraska": "NE",
    "new_hampshire": "NH",
    "new_jersey": "NJ",
    "new_mexico": "NM",
    "nevada": "NV",
    "new_york": "NY",
    "ohio": "OH",
    "oklahoma": "OK",
    "oregon": "OR",
    "pennsylvania": "PA",
    "rhode_island": "RI",
    "south_carolina": "SC",
    "south_dakota": "SD",
    "tennessee": "TN",
    "texas": "TX",
    "utah": "UT",
    "virginia": "VA",
    "vermont": "VT",
    "washington": "WA",
    "wisconsin": "WI",
    "west_virginia": "WV",
    "wyoming": "WY",
}
STATE_ABBR = set(STATE_NAME2ABBR.values())

state_correction = make_flat_wordfinder(STATE_NAME2ABBR.keys())


def state_to_code(name):
    """Get the two letter code from a state name. Return None on failure."""
    try:
        name = name.strip()
    except:
        return None
    if name.upper() in STATE_ABBR:
        return name.upper()
    name = name.lower().replace(" ", "_")
    name = state_correction(name)
    if name is None:
        return None
    if name in STATE_NAME2ABBR:
        return STATE_NAME2ABBR[name]
    else:
        return None


COUNTRY_3LETTER_CODES = {
    "afghanistan": "AFG",
    "aland_islands": "ALA",
    "albania": "ALB",
    "algeria": "DZA",
    "american_samoa": "ASM",
    "andorra": "AND",
    "angola": "AGO",
    "anguilla": "AIA",
    "antarctica": "ATA",
    "antigua_barbuda": "ATG",
    "argentina": "ARG",
    "armenia": "ARM",
    "aruba": "ABW",
    "australia": "AUS",
    "austria": "AUT",
    "azerbaijan": "AZE",
    "bahamas": "BHS",
    "bahrain": "BHR",
    "bangladesh": "BGD",
    "barbados": "BRB",
    "belarus": "BLR",
    "belgium": "BEL",
    "belize": "BLZ",
    "benin": "BEN",
    "bermuda": "BMU",
    "bhutan": "BTN",
    "bolivia": "BOL",
    "bonaire,_sint_eustatius_saba": "BES",
    "bonaire": "BES",
    "bosnia_herzegovina": "BIH",
    "botswana": "BWA",
    "bouvet_island": "BVT",
    "brazil": "BRA",
    "british_indian_ocean_territory": "IOT",
    "brunei_darussalam": "BRN",
    "bulgaria": "BGR",
    "burkina_faso": "BFA",
    "burundi": "BDI",
    "cambodia": "KHM",
    "cameroon": "CMR",
    "canada": "CAN",
    "cape_verde": "CPV",
    "cayman_islands": "CYM",
    "central_african_republic": "CAF",
    "chad": "TCD",
    "chile": "CHL",
    "china": "CHN",
    "prc": "CHN",
    "christmas_island": "CXR",
    "cocos_islands": "CCK",
    "cocos_(keeling)_islands": "CCK",
    "colombia": "COL",
    "comoros": "COM",
    "congo": "COG",
    "republic_congo": "COG",
    "congo,_democratic_republic": "COD",
    "democratic_republic_congo": "COD",
    "dr_congo": "COD",
    "drc": "COD",
    "droc": "COD",
    "cook_islands": "COK",
    "costa_rica": "CRI",
    "cote_d'ivoire": "CIV",
    "cote_divoire": "CIV",
    "croatia": "HRV",
    "cuba": "CUB",
    "curaçao": "CUW",
    "cyprus": "CYP",
    "czechia": "CZE",
    "czech_republic": "CZE",
    "czechoslovakia": None,  # dissolved in 1993
    "denmark": "DNK",
    "djibouti": "DJI",
    "dominica": "DMA",
    "dominican_republic": "DOM",
    "ecuador": "ECU",
    "egypt": "EGY",
    "el_salvador": "SLV",
    "equatorial_guinea": "GNQ",
    "eritrea": "ERI",
    "estonia": "EST",
    "ethiopia": "ETH",
    "europe": None,
    "falkland_islands_(malvinas)": "FLK",
    "falkland_islands": "FLK",
    "faroe_islands": "FRO",
    "fiji": "FJI",
    "finland": "FIN",
    "france": "FRA",
    "french_guiana": "GUF",
    "french_polynesia": "PYF",
    "french_southern_territories": "ATF",
    "gabon": "GAB",
    "gambia": "GMB",
    "gaza_strip": None,
    "georgia": "GEO",
    "germany": "DEU",
    "ghana": "GHA",
    "gibraltar": "GIB",
    "greece": "GRC",
    "greenland": "GRL",
    "grenada": "GRD",
    "guadeloupe": "GLP",
    "guam": "GUM",
    "guatemala": "GTM",
    "guernsey": "GGY",
    "guinea": "GIN",
    "guinea-bissau": "GNB",
    "guyana": "GUY",
    "haiti": "HTI",
    "heard_mc_donald_islands": "HMD",
    "holy_see_(vatican_city_state)": "VAT",
    "vatican_city_state": "VAT",
    "vatican_city": "VAT",
    "honduras": "HND",
    "hong_kong": "HKG",
    "hungary": "HUN",
    "iceland": "ISL",
    "india": "IND",
    "indonesia": "IDN",
    "iran": "IRN",
    "iran,_islamic_republic": "IRN",
    "islamic_republic_iran": "IRN",
    "iraq": "IRQ",
    "ireland": "IRL",
    "isle_man": "IMN",
    "israel": "ISR",
    "italy": "ITA",
    "jamaica": "JAM",
    "japan": "JPN",
    "jersey": "JEY",
    "jordan": "JOR",
    "kazakhstan": "KAZ",
    "kenya": "KEN",
    "kiribati": "KIR",
    "north_korea": "PRK",
    "korea,_democratic_people's_republic": "PRK",
    "democratic_people's_republic_korea": "PRK",
    "south_korea": "KOR",
    "korea": "KOR",
    "republic_korea": "KOR",
    "kosovo": None,
    "kuwait": "KWT",
    "kyrgyzstan": "KGZ",
    "lab": None,
    "laos": "LAO",
    "lao_people's_democratic_republic": "LAO",
    "latvia": "LVA",
    "lebanon": "LBN",
    "lesotho": "LSO",
    "liberia": "LBR",
    "libya": "LBY",
    "state_libya": "LBY",
    "liechtenstein": "LIE",
    "lithuania": "LTU",
    "luxembourg": "LUX",
    "macao": "MAC",
    "macau": "MAC",
    "north_macedonia": "MKD",  # as of Feb 2019 through the ratification of the Prespa agreement,
    "madagascar": "MDG",
    "malawi": "MWI",
    "malaysia": "MYS",
    "maldives": "MDV",
    "mali": "MLI",
    "malta": "MLT",
    "marshall_islands": "MHL",
    "martinique": "MTQ",
    "mauritania": "MRT",
    "mauritius": "MUS",
    "mayotte": "MYT",
    "mexico": "MEX",
    "micronesia,_federated_states": "FSM",
    "federated_states_micronesia": "FSM",
    "micronesia": "FSM",
    "middle_east": None,
    "moldova": "MDA",
    "republic_moldova": "MDA",
    "monaco": "MCO",
    "mongolia": "MNG",
    "montenegro": "MNE",
    "montserrat": "MSR",
    "morocco": "MAR",
    "mozambique": "MOZ",
    "myanmar": "MMR",
    "namibia": "NAM",
    "nauru": "NRU",
    "nepal": "NPL",
    "netherlands": "NLD",
    "new_caledonia": "NCL",
    "new_zealand": "NZL",
    "nicaragua": "NIC",
    "niger": "NER",
    "nigeria": "NGA",
    "niue": "NIU",
    "norfolk_island": "NFK",
    "northern_mariana_islands": "MNP",
    "norway": "NOR",
    "oman": "OMN",
    "pakistan": "PAK",
    "palau": "PLW",
    "palestinian_territory,_occupied": "PSE",
    "occupied_palestinian_territory": "PSE",
    "palestine": "PSE",
    "state_palestine": "PSE",
    "panama": "PAN",
    "papua_new_guinea": "PNG",
    "paraguay": "PRY",
    "peru": "PER",
    "philippines": "PHL",
    "pitcairn": "PCN",
    "poland": "POL",
    "portugal": "PRT",
    "puerto_rico": "PRI",
    "qatar": "QAT",
    "republic_serbia": "SRB",
    "serbia": "SRB",
    "reunion": "REU",
    "romania": "ROU",
    "russia_federation": "RUS",
    "russia": "RUS",
    "ussr": "RUS",  # give or take a few countries,
    "rwanda": "RWA",
    "saint_barthélemy": "BLM",
    "saint_helena": "SHN",
    "saint_kitts_nevis": "KNA",
    "saint_lucia": "LCA",
    "saint_martin": "MAF",
    "saint_pierre_miquelon": "SPM",
    "saint_vincent_grenadines": "VCT",
    "samoa": "WSM",
    "san_marino": "SMR",
    "sao_tome_principe": "STP",
    "saudi_arabia": "SAU",
    "senegal": "SEN",
    "seychelles": "SYC",
    "sierra_leone": "SLE",
    "singapore": "SGP",
    "sint_maarten": "SXM",
    "slovakia": "SVK",
    "slovenia": "SVN",
    "solomon_islands": "SLB",
    "somalia": "SOM",
    "south_africa": "ZAF",
    "south_georgia_south_sandwich_islands": "SGS",
    "south_sudan": "SSD",
    "spain": "ESP",
    "sri_lanka": "LKA",
    "sudan": "SDN",
    "suriname": "SUR",
    "svalbard_jan_mayen": "SJM",
    "swaziland": "SWZ",
    "sweden": "SWE",
    "switzerland": "CHE",
    "syrian_arab_republic": "SYR",
    "syria": "SYR",
    "taiwan": "TWN",
    "tajikistan": "TJK",
    "tanzania,_united_republic": "TZA",
    "thailand": "THA",
    "timor-leste": "TLS",
    "togo": "TGO",
    "tokelau": "TKL",
    "tonga": "TON",
    "trinidad_tobago": "TTO",
    "tunisia": "TUN",
    "turkey": "TUR",
    "northern_cyprus": None,  # No, I don't recognize you either
    "turkmenistan": "TKM",
    "turks_caicos_islands": "TCA",
    "tuvalu": "TUV",
    "uganda": "UGA",
    "ukraine": "UKR",
    "united_arab_emirates": "ARE",
    "united_kingdom": "GBR",
    "united_kingdom_of_great_britain": "GBR",
    "uk": "GBR",
    "britain": "GBR",
    "england": "GBR",
    "united_states_minor_outlying_islands": "UMI",
    "united_states": "USA",
    "united_states_america": "USA",
    "us": "USA",
    "america": "USA",
    "gringoland": "USA",
    "uruguay": "URY",
    "uzbekistan": "UZB",
    "vanuatu": "VUT",
    "venezuela": "VEN",
    "viet_nam": "VNM",
    "vietnam": "VNM",
    "virgin_islands,_british": "VGB",
    "british_virgin_islands": "VGB",
    "virgin_islands,_u.s.": "VIR",
    "u.s._virgin_islands": "VIR",
    "united_states_virgin_islands": "VIR",
    "us_virgin_islands": "VIR",
    "american_virgin_islands": "VIR",
    "wallis_futuna": "WLF",
    "west_bank": None,
    "western_sahara": "ESH",
    "yemen": "YEM",
    "zambia": "ZMB",
    "zimbabwe": "ZWE",
}
COUNTRY_NAMES = set(COUNTRY_3LETTER_CODES.keys())
COUNTRY_ABBREVIATIONS = set(COUNTRY_3LETTER_CODES.values())

clean_name = re.compile("of_|the_|and_|_of|_the|_and")

country_correction = make_flat_wordfinder(COUNTRY_NAMES)


def country_to_code(name):
    """Get the ISO 3-letter codes for a country. Return None on failure."""
    try:
        name = name.strip()
    except:
        return None
    if name.upper() in COUNTRY_ABBREVIATIONS:
        return name.upper()
    name = name.lower().strip().replace(" ", "_")
    name = clean_name.sub("", name)
    name = country_correction(name)
    if name is None:
        return None
    if name in COUNTRY_NAMES:
        return COUNTRY_3LETTER_CODES[name]
    else:
        return None


LOCATIONS_IN = {
    # Canadian territories
    "ontario": "CAN",
    "quebec": "CAN",
    "nova_scotia": "CAN",
    "new_brunswick": "CAN",
    "manitoba": "CAN",
    "british_columbia": "CAN",
    "prince_edward_island": "CAN",
    "saskatchewan": "CAN",
    "alberta": "CAN",
    "newfoundland": "CAN",
    "newfoundland_and_labrador": "CAN",
    # Chinese provinces
    "anhui": "CHN",
    "beijing": "CHN",
    "chongqing": "CHN",
    "fujian": "CHN",
    "guangdong": "CHN",
    "gansu": "CHN",
    "guangxi": "CHN",
    "guizhou": "CHN",
    "henan": "CHN",
    "hubei": "CHN",
    "hebei": "CHN",
    "hainan": "CHN",
    "heilongjiang": "CHN",
    "hong_kong": "CHN",
    "hunan": "CHN",
    "jilin": "CHN",
    "jiangsu": "CHN",
    "jiangxi": "CHN",
    "liaoning": "CHN",
    "macau": "CHN",
    "qinghai": "CHN",
    "sichuan": "CHN",
    "shandong": "CHN",
    "shanghai": "CHN",
    "shaanxi": "CHN",
    "shanxi": "CHN",
    "tianjin": "CHN",
    "tibet": "CHN",
    "yunnan": "CHN",
    "zhejiang": "CHN",
}


def location_to_country_code(name):
    """Map locations to countries"""
    code = country_to_code(name)
    if not code:
        try:
            code = LOCATIONS_IN[name.strip().lower()]
        except:
            if state_to_code(name):
                code = "USA"
            else:
                code = None
    return code



import parsec as p
import re
from octofludb.domain.date import p_year
from octofludb.domain.identifier import p_A0
from octofludb.parser import wordset
from octofludb.util import rmNone, concat

# the 8 segments of the flu genome (order matters)
SEGMENT = ["PB2", "PB1", "PA", "HA", "NP", "NA", "M", "NS"]

p_HA = p.regex("H\d+") ^ p.regex("pdmH\d+")
p_NA = p.regex("N\d+") ^ p.regex("N\d+pdm")

p_ns = p.regex("NS1?").parsecmap(lambda x: "NS")
p_m = p.regex("M[P1]?").parsecmap(lambda x: "M")
p_internal_gene = p.regex("PB2|PB1|PA|NP") ^ p_ns ^ p_m

p_segment = p_internal_gene ^ p.string("HA") ^ p.string("NA")
p_constellation = p.regex("[TPVHX-]{6}|MIXED")
p_segment_number = p.regex("[1-8]")
p_segment_subtype = p_segment ^ p_HA ^ p_NA


@p.generate
def p_subtype_unmixed():
    yield p.regex("(A *\/ *)?")
    ha = yield p_HA
    host = yield p.regex("(hu|sw|av)?")
    na = yield p_NA
    v = yield p.regex("(v)?")
    return ha + host + na + v


p_subtype_mixed = p.regex("mixed", re.I).parsecmap(lambda x: "mixed")
p_subtype = p_subtype_mixed ^ p_subtype_unmixed


def mapreplace(x, pattern, replace):
    if x == pattern:
        return replace
    else:
        return x


p_h1_clade = wordset(
    [
        "alpha",
        "beta",
        "delta1",
        "delta1a",
        "delta1b",
        "delta2",
        "gamma",
        "gamma2",
        "gamma2-beta-like",
        "gamma2_beta_like",
        "pandemic",
        "pdm",
        "pdmH1",
        "human-delta",
        "huVac",
        "predelta",
    ],
    label="h1_clade",
)
p_h3_clade = wordset(
    [
        "2010.1",
        "2010.2",
        "Cluster_I",
        "Cluster_II",
        "Cluster_III",
        "Cluster_IV",
        "Cluster_IVA",
        "Cluster_IVB",
        "Cluster_IVC",
        "Cluster_IVD",
        "Cluster_IVE",
        "Cluster_IVF",
        "I",
        "II",
        "III",
        "IV",
        "IV-A",
        "IV-B",
        "IV-C",
        "IV-D",
        "IV-E",
        "IV-F",
        "huVac",
        "human-like_2010.1",
        "human-like_2010.2",
        "human-like_2016",
    ],
    label="h3_clade",
)
p_n1_clade = wordset(
    ["Human_seasonal", "huVac", "Classical", "Pandemic", "MN99"], label="n1_clade"
)
p_n2_clade = wordset(
    [
        "Human_N2",
        "2016",
        "Human-like",
        "1998",
        "1998A",
        "98A",
        "98A1",
        "98A_1",
        "98A2",
        "98A_2",
        "1998B",
        "98B",
        "98B1",
        "98B_1",
        "98B2",
        "98B_2",
        "2002",
        "2002A",
        "02A1",
        "02A2",
        "2002B",
        "02B1",
        "02B2",
        "TX98",
    ],
    label="n3_clade",
)

p_internal_gene_clade = wordset(
    ["PDM", "TRIG", "LAIV"], label="internal_gene_clade"
).parsecmap(lambda x: x.upper())


class Strain:
    def __init__(
        self,
        flutype=None,
        subtype=None,
        host=None,
        place=None,
        ident=None,
        year=None,
        raw=None,
    ):
        self.flutype = flutype
        self.host = host
        self.place = place
        self.ident = ident
        self.year = year
        self.subtype = subtype
        try:
            self.a0 = p_A0.parse(self.ident)
        except:
            self.a0 = None

    def getHost(self):
        if self.host:
            return self.host
        else:
            return "human"

    def __str__(self):
        fields = ["A", self.host, self.place, self.ident, self.year]
        return "/".join(rmNone(fields))


p_strain_field = p.regex("[^|/\n\t]+")

# A/Alaska/1935
# A/Alaska/1935(H1N1)
# A/New Jersey/1940
# A/Denver/1957
@p.generate
def p_s3():
    flutype = yield p.regex("[ABCD]") << p.string("/")
    place = yield p_strain_field << p.string("/")
    year = yield p_year
    subtype = yield p.optional(p.string("(") >> p_subtype << p.string(")"))
    return Strain(flutype=flutype, place=place, year=year, subtype=subtype)


# A/Baylor/11735/1982
# A/Berkeley/1/66
# A/California/NHRC-OID_SAR10587N/2018
# A/District of Columbia/WRAIR1753P/2010(H3N2)
# A/District_Of_Columbia/03/2014
@p.generate
def p_s4():
    flutype = yield p.regex("[ABCD]") << p.string("/")
    place = yield p_strain_field << p.string("/")
    ident = yield p_strain_field << p.string("/")
    year = yield p_year
    subtype = yield p.optional(p.string("(") >> p_subtype << p.string(")"))
    return Strain(flutype=flutype, place=place, year=year, ident=ident, subtype=subtype)


# A/Swine/Iowa/533/99
# A/swine/Iowa/3421/1990
# A/swine/Nebraska/00722/2005_mixed_
# A/swine/Ontario/55383/04
# A/swine/Oklahoma/A01785279/2017
@p.generate
def p_s5():
    flutype = yield p.regex("[ABCD]") << p.string("/")
    host = yield p_strain_field << p.string("/")
    place = yield p_strain_field << p.string("/")
    ident = yield p_strain_field << p.string("/")
    year = yield p_year
    subtype = yield p.optional(p.string("(") >> p_subtype << p.string(")"))
    return Strain(
        flutype=flutype, host=host, place=place, year=year, ident=ident, subtype=subtype
    )


p_strain_obj = p_s5 ^ p_s4 ^ p_s3



import parsec as p
from octofludb.util import padDigit, rmNone
from octofludb.parser import wordset
from rdflib.namespace import XSD
from rdflib import Literal


def expandYear(x: str) -> str:
    """Expand years: [1-9]X -> 19XX and 0X -> 200X"""
    if len(x) == 2:
        if int(x[0]) <= 2:
            x = "20" + x  # 00-29 are cast as 2000-2019
        else:
            x = "19" + x  # 30-99 are cast as 1920-1999
    return x


class Date:
    def __init__(self, year: str, month: str = None, day: str = None):
        self.year = year
        self.month = month
        self.day = day

    def as_uri(self):
        # 2015
        if self.year and self.month is None:
            uri = Literal(self.year, datatype=XSD.gYear)
        # 2015/05
        elif self.year and self.month and self.day is None:
            uri = Literal(f"{self.year}-{self.month}", datatype=XSD.gYearMonth)
        # 2015/05/31
        elif self.year and self.month and self.day:
            uri = Literal(f"{self.year}-{self.month}-{self.day}", datatype=XSD.date)
        # 05/31
        elif self.year is None and self.month and self.day:
            uri = Literal(f"{self.month}-{self.day}", datatype=XSD.gMonthDay)
        # 05
        elif self.year is None and self.month and self.day is None:
            uri = Literal(f"{self.month}", datatype=XSD.gMonth)
        # 31
        elif self.year is None and self.month is None and self.day:
            uri = Literal(f"{self.day}", datatype=XSD.gDay)
        else:
            uri = None
        return uri

    def __str__(self):
        return "-".join(rmNone([self.year, self.month, self.day]))


@p.generate
def p_date_ymd():
    y = yield p_longyear
    yield p.optional(p.regex("[-/]"))
    m = yield p_month
    yield p.optional(p.regex("[-/]"))
    d = yield p_day
    yield p.optional(p.regex(" \d\d:\d\d:\d\d(\.\d+)?"))
    return Date(month=m, day=d, year=y)


@p.generate
def p_date_mdy():
    m = yield p_month
    yield p.optional(p.regex("[-/]"))
    d = yield p_day
    yield p.optional(p.regex("[-/]"))
    y = yield p_longyear
    yield p.optional(p.regex(" \d\d:\d\d:\d\d(\.\d+)?"))
    return Date(month=m, day=d, year=y)


@p.generate
def p_date_dMy():
    """
    01-Apr-2002
    """
    d = yield p_day
    yield p.optional(p.regex("[-/]"))
    m = yield p_month_str
    yield p.optional(p.regex("[-/]"))
    y = yield p_year
    yield p.optional(p.regex(" \d\d:\d\d:\d\d(\.\d+)?"))
    return Date(month=m, day=d, year=y)


@p.generate
def p_date_polite():
    """
    May 31, 2018
    """
    m = yield p_month_str
    yield p.spaces()
    d = yield p_day
    yield p.string(",")
    yield p.spaces()
    y = yield p_longyear
    return Date(day=d, month=m, year=y)


@p.generate
def p_date_my():
    m = yield p_month
    yield p.regex("[-/]")
    y = yield p_longyear
    return Date(month=m, year=y)


@p.generate
def p_date_ym():
    y = yield p_longyear
    yield p.regex("[-/]")
    m = yield p_month
    return Date(month=m, year=y)


@p.generate
def p_utc_date():
    y = yield p_longyear
    yield p.optional(p.string("-"))
    m = yield p_month_num
    yield p.optional(p.string("-"))
    d = yield p_day
    yield p.string("T")
    yield p_iso8601_time
    return Date(year=y, month=m, day=d)


@p.generate
def p_iso8601_time():
    h = yield p_hour
    yield p.optional(p.string(":"))
    m = yield p_minute
    yield p.optional(p.string(":"))
    s = yield p_second
    yield p.optional((p.string("Z") | (p.string("+") >> p.regex("\d\d:\d\d"))))
    return h + m + s


p_hour = p.regex("[01]\d") ^ p.regex("2[0-3]")
p_minute = p.regex("[0-5]\d")
p_second = p.regex("[0-5]\d")


p_year = p.regex("20\d\d") ^ p.regex("1\d\d\d") ^ p.regex("\d\d").parsecmap(expandYear)
p_longyear = p.regex("20\d\d") | p.regex("1[89]\d\d")
p_day = p.regex("3[01]|[012]?\d").parsecmap(padDigit)

months = {
    "jan": "1",
    "feb": "2",
    "mar": "3",
    "apr": "4",
    "may": "5",
    "jun": "6",
    "jul": "7",
    "aug": "8",
    "sep": "9",
    "oct": "10",
    "nov": "11",
    "dec": "12",
    "january": "1 ",
    "february": "2",
    "march": "3",
    "april": "4",
    "may": "5",
    "june": "6",
    "july": "7",
    "august": "8",
    "september": "9",
    "october": "10",
    "november": "11",
    "december": "12",
}

p_month_num = p.regex("10|11|12|0?[1-9]").parsecmap(padDigit)
p_month_str = (
    wordset(months.keys(), "date")
    .parsecmap(lambda x: months[x.lower()])
    .parsecmap(padDigit)
)
p_month = p_month_num ^ p_month_str

p_date = p_utc_date ^ p_date_polite ^ p_date_ymd ^ p_date_mdy ^ p_date_dMy

p_any_date = (
    p_utc_date
    ^ p_date_polite
    ^ p_date_dMy
    ^ p_date_ymd
    ^ p_date_mdy
    ^ p_date_my
    ^ p_date_ym
    ^ p_year.parsecmap(lambda y: Date(year=y))
)



from octofludb.nomenclature import (
    uidgen,
    P,
    O,
    make_property,
    make_uri,
    make_date,
    make_country_uri,
    make_usa_state_uri,
)
from octofludb.util import rmNone, make_maybe_add
from octofludb.hash import chksum
from rdflib import Literal
from rdflib.namespace import XSD
import re
import octofludb.domain.identifier as identifier
import octofludb.domain.flu as flu
import octofludb.domain.animal as animal
import sys
from octofludb.util import log
from octofludb.colors import bad
import octofludb.domain.geography as geo


def add_gb_meta_triples(g, gb_meta, only_influenza_a=True):
    """
    Add genbank triples

    Parameters
    ----------
    g : rdflib Graph object
    gb_meta : Genbank metadata dictionary from Bio.Entrez
    only_influenza_a : bool

    Returns
    -------
    string containing any raised error or warning message
    """

    error_entry = ""

    try:
        accession = str(gb_meta["GBSeq_primary-accession"])
    except:
        log(bad("Bad Genbank Entry"))
        return "Unknown\tNo primary accession"

    if only_influenza_a:
        # ignore this entry if the organism is not specified
        if not "GBSeq_organism" in gb_meta:
            return f"{accession}\tNo organsim specified"
        # ignore this entry if the organism is not an Influenza virus
        if not bool(re.match("Influenza [ABCD] virus", gb_meta["GBSeq_organism"])):
            return f"{accession}\tNot influenza"

    gid = make_uri(accession)
    g.add((gid, P.gb, Literal(accession)))

    maybe_add = make_maybe_add(g, gb_meta, gid)

    def make_integer(x):
        return rdflib.Literal(x, datatype=XSD.integer)

    maybe_add(P.gb_locus, "GBSeq_locus")
    maybe_add(P.gb_length, "GBSeq_length", formatter=make_integer)
    maybe_add(P.gb_strandedness, "GBSeq_strandedness")
    maybe_add(P.gb_moltype, "GBSeq_moltype")
    maybe_add(P.gb_topology, "GBSeq_topology")
    maybe_add(P.gb_division, "GBSeq_division")
    maybe_add(P.gb_update_date, "GBSeq_update-date", formatter=make_date)
    maybe_add(P.gb_create_date, "GBSeq_create-date", formatter=make_date)
    maybe_add(P.gb_definition, "GBSeq_definition")
    maybe_add(P.gb_primary_accession, "GBSeq_primary_accession")
    maybe_add(P.gb_accession_version, "GBSeq_accession-version")
    maybe_add(P.gb_source, "GBSeq_source")
    maybe_add(P.gb_organism, "GBSeq_organism")
    maybe_add(P.gb_taxonomy, "GBSeq_taxonomy")

    # usually an entry has sequence, but there are weird exceptions
    if "GBSeq_sequence" in gb_meta:
        seq = gb_meta["GBSeq_sequence"].upper()
        g.add((gid, P.dnaseq, Literal(seq)))
        g.add((gid, P.chksum, Literal(chksum(seq))))

    strain = None
    host = None
    date = None
    country = None

    igen = uidgen(base=accession + "_feat_")
    for feat in gb_meta["GBSeq_feature-table"]:
        fid = next(igen)
        g.add((gid, P.has_feature, fid))
        g.add((fid, P.name, Literal(feat["GBFeature_key"])))

        maybe_add = make_maybe_add(g, feat, fid)
        maybe_add(P.gb_location, "GBFeature_location")
        #  maybe_add(P.gb_key, "GBFeature_intervals") # for laters

        if "GBFeature_quals" in feat:
            for qual in feat["GBFeature_quals"]:
                if not ("GBQualifier_name" in qual and "GBQualifier_value" in qual):
                    continue
                key = qual["GBQualifier_name"]
                val = qual["GBQualifier_value"]

                if key == "translation":
                    g.add((fid, P.proseq, Literal(val)))
                    g.add((fid, P.chksum, Literal(chksum(val))))
                elif key == "strain":
                    try:
                        strain = identifier.p_strain.parse(val)
                    except:
                        log(bad("Bad strain name: ") + val)
                        error_entry = f"{val}\tBad strain name"
                        strain = val
                elif key == "collection_date":
                    date = make_date(val)
                elif key == "host":
                    host = val
                elif key == "country":
                    country = re.sub(":.*", "", val)
                elif key == "gene":
                    try:
                        segment_name = flu.p_segment.parse_strict(val)
                        # attach the segment_name to the top-level genbank record, not the feature
                        g.add((gid, P.segment_name, Literal(segment_name)))
                    except:
                        pass
                    # attach the original, unparsed gene name to the feature
                    g.add((fid, make_property(key), Literal(val)))
                else:
                    g.add((fid, make_property(key), Literal(val)))

    # link strain information
    if strain:
        sid = make_uri(strain)
        g.add((sid, P.has_segment, gid))
        g.add((sid, P.strain_name, Literal(strain)))
        if host:
            g.add((sid, P.host, Literal(animal.clean_host(host))))
        if date:
            g.add((sid, P.date, date))
        if country:
            code = geo.country_to_code(country)
            country_uri = make_country_uri(country)
            g.add((sid, P.country, country_uri))
            if code is None:
                # if this is an unrecognized country (e.g., Kosovo) then state
                g.add((country_uri, P.name, Literal(country)))
            if code == "USA":
                fields = strain.split("/")
                for field in fields[1:]:
                    # If this looks like a US state, add it
                    code = geo.state_to_code(field)
                    if code:
                        g.add((sid, P.state, make_usa_state_uri(code)))
                    # If this looks like an A0 number, add it
                    try:
                        A0 = identifier.p_A0.parse_strict(field)
                        g.add((sid, P.barcode, Literal(A0)))
                    except:
                        pass
    else:
        locus = gb_meta["GBSeq_locus"]
        log(bad("Missing strain: ") + locus)
        error_entry = f"{locus}\tNo strain name"

    return error_entry



import parsec as p
import re

p_A0 = p.regex("A0\d{7}")
p_tosu = p.regex("\d+TOSU\d+")
p_epi_isolate = p.regex("EPI_ISL_\d+")


def clean_strain(x):
    x = x.strip().replace(" ", "_")
    # strain names may be parenthesized:  (A/Bratislava/6/97 (H3N2))
    x = re.sub("^\((.*)\)$", "\\1", x)
    # remove terminal parentheses
    x = re.sub("_*\(.*\)_*$", "", x)
    # remove terminal braces
    x = re.sub("_*\[.*\]_*$", "", x)
    return x


p_strain_no_paren = p.regex("[ABCD]/[^/()\[\]]+/.+").parsecmap(clean_strain)
p_strain_paren = p.regex("\([ABCD]/[^/()\[\]]+/.+\)").parsecmap(clean_strain)
p_strain = p_strain_paren ^ p_strain_no_paren

p_barcode = p_A0 ^ p_tosu ^ p_epi_isolate ^ p_strain  # e.g. A01104095 or 16TOSU4783
p_gb = p.regex("[A-Z][A-Z]?\d{5,7}")
p_epi_id = p.regex("EPI_?\d\d\d+")
p_seqid = p_gb ^ p_epi_id

p_global_clade = (
    p.regex("\d[ABC]([\._-]\d+){1,4}([_-]?like)?([_-]?vaccine)?")
    ^ p.regex("Other-[A-Za-z]*[0-9.a-zA-Z-]*")
    ^ p.regex("3\.[12][09]\d0\.[0-9.a-zA-Z-]+")
    ^ p.regex("(humanVaccine|Outgroup)")
)



import hashlib
import subprocess
import yaml
import pgraphdb
import smof
import glob
import os
import sys
import shutil
import math
from octofludb.classes import Table
from octofludb.util import log, die
import octofludb.colors as colors


def get_data_file(filename):
    return os.path.join(os.path.dirname(__file__), "data", filename)


def octofludbHome():
    return os.path.join(os.path.expanduser("~"), ".octofludb")


def getDataDir(config):
    datadir = expandpath(config["datadir"])[0]
    if not os.path.exists(datadir):
        os.mkdir(datadir)
    return datadir


def error_log_entry(entries, logfile):
    homedir = octofludbHome()
    logdir = os.path.join(homedir, "logs")
    logpath = os.path.join(logdir, logfile)

    if not os.path.exists(logdir):
        os.mkdir(logdir)

    with open(logpath, "a") as f:
        for line in entries:
            print(line, file=f)

    return logpath


def epiflu_fasta_files(config):
    try:
        data_home = getDataDir(config)
    except KeyError:
        die("The config file is missing a `datadir` entry")
    except IndexError:
        die("The path to the `datadir` entry in config does not exist")
    return expandpath(os.path.join(data_home, config["epiflu_fasta"]))


def epiflu_meta_files(config):
    try:
        data_home = getDataDir(config)
    except KeyError:
        die("The config file is missing a `datadir` entry")
    except IndexError:
        die("The path to the `datadir` entry in config does not exist")
    return expandpath(os.path.join(data_home, config["epiflu_meta"]))


def get_octoflu_reference(config):
    try:
        refpath = config["octoflu_reference"]
        if refpath:
            reference = expandpath(os.path.join(octofludbHome(), refpath))[0]
        else:
            reference = None
    except KeyError:
        die("The config file is missing a `datadir` entry")
    except IndexError:
        die("The path to the `datadir` entry in config does not exist")
    return reference


def tag_files(config, tag):
    try:
        data_home = expandpath(config["datadir"])[0]
    except KeyError:
        die("The config file is missing a `datadir` entry")
    except IndexError:
        die("The path to the `datadir` entry in config does not exist")

    try:
        tagfile = config["tags"][tag]
    except KeyError:
        die(f"Could not find tag {tag} in config")

    return expandpath(os.path.join(data_home, tagfile))


def initialize_config_file():
    """
    Create a default config file is none is present in the octofludb home directory
    """
    config_template_file = os.path.join(
        os.path.dirname(__file__), "data", "config.yaml"
    )
    config_local_file = os.path.join(octofludbHome(), "config.yaml")

    if not os.path.exists(config_local_file):
        print(
            f" - Creating config template at '{str(config_local_file)}'",
            file=sys.stderr,
        )
        shutil.copyfile(config_template_file, config_local_file)

    return config_local_file


def load_config_file():
    """
    Load the local config file (create the config file first if it does not exist)
    """
    config_local_file = initialize_config_file()
    with open(config_local_file, "r") as f:
        try:
            config = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            print(exc, file=sys.stderr)
            sys.exit(1)
    return config


def file_md5sum(path):
    with open(path, "rb") as f:
        file_hash = hashlib.md5(f.read()).hexdigest()
    return file_hash


def evenly_divide(total, preferred_size):
    n = max(math.ceil(total / preferred_size), 1)
    size = total // n
    return [size + (i < total - size * n) for i in range(n)]


def partition(xs, sizes):
    xss = []
    start = 0
    for size in sizes:
        if start >= len(xs):
            break
        else:
            xss.append(xs[start : start + size])
            start += size
    return xss


def runOctoFLU(path, reference=None):
    """
    Run octoFLU on the given fasta paths.

    OctoFLU mangles names terribly, so it is important to ensure that the input
    names are appropriate segment ids (e.g., genbank ids or epiflu ids).
    """

    # The given path may be a glob (e.g., `data/*fna`), so expand to all
    # fasta files and make the paths absolute
    fastafiles = expandpath(path)

    log(fastafiles)

    if reference:
        try:
            reference = expandpath(reference)[0]
        except IndexError:
            die(f"The path {reference} does not point to a file")

    # Store the current working directory so that we can return to it at the
    # end of this function
    cwd = os.getcwd()

    def cleanup():
        os.chdir(cwd)

    # List of all files that have been successfully created. These will be
    # concatenated together and uploaded after a successful run.
    created_files = []

    try:

        # Everything in this build is relative to the default build directory specified in the config file
        gotoBuildHome()

        # Clone the octoFLU repository IF it is not already present (this
        # command doesn't pull the latest version, that is up to you, I guess).
        cloneGithubRepo("flu-crew", "octoFLU")

        # Move to the octoFLU repo directory
        os.chdir("octoFLU")

        # This is the path to the default reference fasta file
        reference_path = os.path.join("reference_data", "reference.fa")

        # if a reference file is given, copy it over and save the original reference
        if reference:
            # copy the original reference file
            os.rename(reference_path, "reference.fa~")
            shutil.copy(reference, reference_path)

        for fastafile in fastafiles:
            # open the fasta file as a list of FastaEntry objects
            fna = list(smof.uniq_headers(smof.open_fasta(fastafile)))

            if len(fna) == 0:
                next

            # break the input fasta into small pieces so we don't kill our tree builder
            for (i, chunk) in enumerate(partition(fna, evenly_divide(len(fna), 5000))):
                # create a default name for the fasta file chunk
                chunk_relpath = f"x{str(i)}_{os.path.basename(fastafile)}"
                with open(chunk_relpath, "w") as chunk_fh:
                    # write the FastaEntry list to the chunk filename
                    smof.print_fasta(chunk, out=chunk_fh)
                    # run octoFLU using the given reference
                    try:
                        subprocess.run(["./octoFLU.sh", chunk_relpath], check=True)
                    except subprocess.CalledProcessError as e:
                        log(colors.bad(f"`./octoFLU.sh {chunk_relpath}` failed"))
                        raise e
                    # if the octoFLU command was successful, it will have created a table in the location below
                    table_path = os.path.join(
                        f"{chunk_relpath}_output", f"{chunk_relpath}_Final_Output.txt"
                    )
                    # add the absolute path to this table to the created file list
                    created_files.append(expandpath(table_path)[0])

        results = []
        for filename in created_files:
            with open(filename, "r") as f:
                results += [
                    [r.strip() for r in line.split("\t")[0:4]] for line in f.readlines()
                ]

        # move the original reference file back if it was moved
        if reference and os.path.exists("reference.fa~"):
            os.rename("reference.fa~", reference_path)
    except Exception as e:
        cleanup()
        log(colors.bad("octoFLU run failed"))
        raise e

    cleanup()

    return results


def findMotifs(sparql_filename, patterns, subtype, url, repo_name):
    import octofludb.formatting as formatting
    import pgraphdb as db
    import flutile

    # write fasta file
    fasta_filename = f"{subtype}.fna"
    results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo_name)
    with open(fasta_filename, "w") as f:
        formatting.write_as_fasta(results, outfile=f)

    # use flutile to find motifs
    motif_filename = f"{subtype}-motif.tab"
    flutile.write_bounds(
        tabular=True,
        motif_strs=patterns,
        keep_signal=False,
        subtype=subtype,
        fasta_file=fasta_filename,
        conversion="dna2aa",
        outfile=motif_filename,
    )

    return motif_filename


def cloneGithubRepo(user, repo):
    """
    Clone a github repository if the repo folder is not already present.
    """
    if not os.path.exists(repo):
        subprocess.run(["git", "clone", f"http://github.com/{user}/{repo}"])


def buildHome():
    return os.path.join(octofludbHome(), "build")


def gotoBuildHome():
    # move to octofludb build home
    build_dir = buildHome()
    if not os.path.exists(build_dir):
        os.mkdir(build_dir)
    print(f"Moving to {build_dir}", file=sys.stderr)
    os.chdir(build_dir)


def expandpath(path):
    """
    Expands globs and gets absolute paths

    This command NEVER fails. If nothing in a path exists, an empty list is returned.
    """
    return glob.glob(os.path.abspath(os.path.expanduser(path)))



from termcolor import colored


def good(text):
    coloredText = colored(text, "green", attrs=["bold"])
    return coloredText


def bad(text):
    coloredText = colored(text, "red", attrs=["bold"])
    return coloredText



import parsec as p
from octofludb.hash import chksum
from octofludb.util import log
from octofludb.domain.flu import SEGMENT
from rdflib import Literal
from collections import OrderedDict
import re

from octofludb.domain.identifier import (
    p_global_clade,
    p_A0,
    p_tosu,
    p_epi_isolate,
    p_strain,
    p_gb,
    p_epi_id,
)

from octofludb.domain.flu import (
    p_HA,
    p_NA,
    p_internal_gene,
    p_segment,
    p_segment_subtype,
    p_segment_number,
    p_subtype,
    p_strain_obj,
    p_constellation,
    p_h1_clade,
    p_h3_clade,
    p_internal_gene_clade,
    p_n1_clade,
    p_n2_clade,
)

from octofludb.domain.date import p_year, p_longyear, p_month, p_day, p_date, p_any_date
from octofludb.domain.geography import (
    country_to_code,
    state_to_code,
    location_to_country_code,
)
from octofludb.domain.animal import p_host
from octofludb.domain.sequence import p_dnaseq, p_proseq

from octofludb.token import Token, Unknown
from octofludb.util import rmNone
from octofludb.nomenclature import (
    make_uri,
    make_date,
    make_property,
    make_usa_state_uri,
    make_literal,
    make_country_uri,
    P,
)

BARCODE_PAT = re.compile("A0\d{7}|\d+TOSU\d+")


class Country(Token):
    typename = "country"
    class_predicate = P.country

    def munge(self, text):
        return country_to_code(text)

    @classmethod
    def testOne(cls, item, na_str=[None]):
        if item in na_str:
            return None
        return country_to_code(item)

    def as_uri(self):
        return make_country_uri(self.dirty)

    def object_of(self, g, uri):
        # I am allowing a link even when there was no match. This allows
        # unusual country names to be treated as countries when the context
        # suggests they are countries. But they do at least have to be
        # non-empty strings.
        if uri and self.dirty:
            g.add((uri, self.as_predicate(), self.as_uri()))


class CountryOrState(Token):
    """
    Maps countries, states, or major cities to the country code
    """

    typename = "country"
    class_predicate = P.country

    def munge(self, text):
        return location_to_country_code(text)

    @classmethod
    def testOne(cls, item, na_str=[None]):
        if item in na_str:
            return None
        return location_to_country_code(item)

    def as_uri(self):
        return make_country_uri_from_code(self.clean)

    def object_of(self, g, uri):
        if uri and self.dirty:
            g.add((uri, self.as_predicate(), self.as_uri()))


class StateUSA(Token):
    typename = "state"
    class_predicate = P.state
    parser = state_to_code

    @classmethod
    def testOne(cls, item, na_str=[None]):
        if item in na_str:
            return None
        return state_to_code(item)

    def object_of(self, g, uri):
        if uri and self.matches:
            g.add((uri, P.state, make_usa_state_uri(self.clean)))


class Date(Token):
    typename = "date"
    parser = p_any_date
    class_predicate = P.date

    def munge(self, text):
        return str(text)

    def as_literal(self):
        return make_date(self.dirty)


class Host(Token):
    typename = "host"
    parser = p_host

    def munge(self, text):
        return text.lower()


STRAIN_FIELDS = {
    "date",
    "submission_date",
    "collection_date",
    "country",
    "state",
    "host",
    "global_clade",
    "subtype",
    "barcode",
    "strain_name",
    "gisaid_strain_name",
}

# --- strain tokens ---
class StrainToken(Token):
    group = "strain"

    def as_uri(self):
        return make_uri(self.clean)

    def munge(self, text):
        return text.replace(" ", "_")

    def _has_segment(self, tokens):
        for token in tokens:
            if token.group == "segment" or token.typename == "dnaseq":
                return True
        return None

    def relate(self, tokens, g, levels=None):
        if self.clean is None or not self.matches:
            return
        uri = self.as_uri()
        has_segment = self._has_segment(tokens)
        use_segment = (levels is None and has_segment) or (
            levels is not None and "segment" in levels and has_segment
        )
        g.add((uri, make_property(self.typename), self.as_literal()))
        for other in tokens:
            if not other.matches or other.clean == self.clean or other.clean is None:
                continue
            if other.group == "segment":
                g.add((uri, P.has_segment, other.as_uri()))
            elif other.choose_field_name() in STRAIN_FIELDS:
                other.object_of(g, uri)
            elif not use_segment:
                other.object_of(g, uri)


class Isolate(StrainToken):
    typename = "isolate_id"
    parser = p_epi_isolate

    def munge(self, text):
        return text.upper()


class Barcode(StrainToken):
    typename = "barcode"
    parser = p_tosu ^ p_A0

    def munge(self, text):
        return text.upper()

    def add_triples(self, g):
        if self.clean:
            g.add((self.as_uri(), P.barcode, self.as_literal()))


class Strain(StrainToken):
    typename = "strain_name"
    parser = p_strain

    def munge(self, text):
        return text.replace(" ", "_")

    def add_triples(self, g):
        if self.clean:
            uri = self.as_uri()
            g.add((uri, P.strain_name, self.as_literal()))
            for el in self.clean.split("/"):
                barcode_match = re.fullmatch(BARCODE_PAT, el)
                state_str = StateUSA.parser(el)

                # some strain names contain a barcode, which is also a unique id
                if barcode_match is not None:
                    g.add((uri, P.strain_name, self.as_literal()))
                    g.add((uri, P.barcode, make_literal(barcode_match[0])))
                elif state_str is not None:
                    state = StateUSA(state_str)
                    state.object_of(g, uri)


# --- strain attributes ---
class StrainAttribute(Token):
    def relate(self, tokens, g, levels=None):
        for other in tokens:
            if other.group == "strain" and other.typename != self.typename:
                self.object_of(g, other.as_uri())


class Subtype(StrainAttribute):
    typename = "subtype"
    parser = p_subtype
    class_predicate = P.subtype


class Constellation(StrainAttribute):
    typename = "constellation"
    parser = p_constellation
    class_predicate = P.constellation


class GlobalClade(StrainAttribute):
    typename = "global_clade"
    parser = p_global_clade
    class_predicate = P.global_clade


class HA(StrainAttribute):
    typename = "HA"
    parser = p_HA
    class_predicate = P.ha_clade


class NA(StrainAttribute):
    typename = "NA"
    parser = p_NA
    class_predicate = P.na_clade


class InternalGene(StrainAttribute):
    typename = "internal_gene"
    parser = p_internal_gene


# --- strain tokens ---
class SegmentToken(Token):
    group = "segment"
    class_predicate = P.has_segment

    def as_uri(self):
        return make_uri(self.clean)

    def relate(self, tokens, g, levels=None):
        if not self.matches:
            return
        uri = self.as_uri()
        for other in tokens:
            if other.clean is None:
                continue
            if (
                other.matches
                and other.group == "segment"
                and other.typename != self.typename
            ):
                g.add((uri, P.sameAs, other.as_uri()))
            elif (
                not other.choose_field_name() in STRAIN_FIELDS
                and other.typename is not None
            ):
                other.object_of(g, uri)


class Genbank(SegmentToken):
    typename = "genbank_id"
    parser = p_gb

    def munge(self, text):
        return text.upper()

    def add_triples(self, g):
        if self.clean:
            g.add((self.as_uri(), P.gb, self.as_literal()))


class EpiSeqid(SegmentToken):
    typename = "epi_id"
    parser = p_epi_id

    def as_uri(self):
        return make_uri(self.clean)

    def munge(self, text):
        return text.upper().replace("_", "")

    def add_triples(self, g):
        if self.clean:
            g.add((self.as_uri(), P.epi_id, self.as_literal()))


# --- segment attributes ---
class SegmentAttribute(Token):
    def relate(self, tokens, g, levels=None):
        for other in tokens:
            if other.group == "segment":
                self.object_of(g, other.as_uri())


class SegmentName(SegmentAttribute):
    typename = "segment_name"
    parser = p_segment


class SegmentSubtype(SegmentAttribute):
    typename = "segment_subtype"
    parser = p_segment_subtype


class SegmentNumber(SegmentAttribute):
    typename = "segment_number"
    parser = p_segment_number

    def object_of(self, g, uri):
        if uri and self.matches:
            g.add((uri, P.segment_number, Literal(self.clean)))
            g.add((uri, P.segment_name, Literal(SEGMENT[int(self.clean) - 1])))


class SequenceToken(Token):
    group = "sequence"

    def munge(self, text):
        return re.sub("[^A-Z*]", "", text.upper())

    def as_uri(self):
        return make_uri(chksum(self.clean))

    def _has_segment(self, tokens):
        for token in tokens:
            if token.group == "segment":
                return True
        return None

    @classmethod
    def goodness(cls, items, na_str=[None]):
        matches = [
            bool(cls.testOne(item=x, na_str=na_str)) and len(str(x)) > 20
            for x in items
            if x not in na_str
        ]
        if len(items) > 0:
            goodness = sum(matches) / len(items)
        else:
            goodness = 0
        return goodness


class Dnaseq(SequenceToken):
    typename = "dnaseq"
    parser = p_dnaseq

    def object_of(self, g, uri):
        if uri and self.matches:
            g.add((uri, P.chksum, Literal(chksum(self.clean))))
            g.add((uri, P.dnaseq, Literal(self.clean)))

    def relate(self, tokens, g, levels=None):
        uri = self.as_uri()
        for other in tokens:
            if other.clean is None:
                continue
            elif other.group == "strain":
                g.add((other.as_uri(), P.has_segment, uri))
            elif not self._has_segment(tokens) and not other.typename in STRAIN_FIELDS:
                other.object_of(g, uri)


class Proseq(SequenceToken):
    typename = "proseq"
    parser = p_proseq

    def relate(self, tokens, g, levels=None):
        uri = self.as_uri()
        g.add((uri, P.proseq, Literal(self.clean)))
        has_segment = self._has_segment(tokens)
        for other in tokens:
            if other.clean is None:
                continue
            if other.group == "segment":
                g.add((other.as_uri(), P.has_feature, uri))
            elif other.group == "strain":
                if has_segment:
                    log("WARNING: I don't know how to connect a protein to a strain id")
            elif not other.choose_field_name() in STRAIN_FIELDS and not has_segment:
                other.object_of(g, uri)


class H1Clade(Token):
    typename = "h1_clade"
    parser = p_h1_clade


class H3Clade(Token):
    typename = "h3_clade"
    parser = p_h3_clade


class US_Clade(Token):
    typename = "us_clade"
    parser = p_h1_clade ^ p_h3_clade


class N1Clade(Token):
    typename = "n1_clade"
    parser = p_n1_clade


class N2Clade(Token):
    typename = "n2_clade"
    parser = p_n2_clade


class InternalGeneClade(Token):
    typename = "internal_gene_clade"
    parser = p_internal_gene_clade


allClassifiers = OrderedDict(
    [
        (c.typename, c)
        for c in [
            Isolate,
            Genbank,
            Barcode,
            Constellation,
            Country,
            Date,
            EpiSeqid,
            GlobalClade,
            Subtype,
            SegmentName,
            SegmentSubtype,
            Host,
            InternalGene,
            SegmentNumber,
            Strain,
            StateUSA,
            InternalGeneClade,
            H1Clade,
            H3Clade,
            US_Clade,
            N1Clade,
            N2Clade,
            Dnaseq,
            Proseq,
            Unknown,
        ]
    ]
)



import sys
from rdflib import Literal
import pandas as pd
import octofludb.classes as classes
from octofludb.colors import bad
import octofludb.classifiers.flucrew as flu
import octofludb.token as tok
import octofludb.domain.identifier as identifier
import parsec
from octofludb.nomenclature import P, O, make_uri, make_tag_uri
from octofludb.util import log, file_str
import re
import math
from tqdm import tqdm
import datetime as datetime


def mk_blast(g, filehandle, tag=None):
    timestr = datetime.datetime.now()
    for row in tqdm(filehandle.readlines()):
        try:
            (
                qseqid,
                sseqid,
                pident,
                length,
                mismatch,
                gapopen,
                qstart,
                qend,
                sstart,
                send,
                evalue,
                bitscore,
            ) = row.split("\t")
        except ValueError:
            sys.exit(
                "Expected blast file to have exactly 12 fields (as per default blast outfmt 6 options)"
            )

        huid = make_uri(f"blast/{qseqid}-{sseqid}-{bitscore}")

        if tag:
            taguri = make_tag_uri(tag)
            g.add((taguri, P.name, Literal(tag)))
            g.add((taguri, P.time, Literal(timestr)))
            g.add((taguri, P.file, Literal(file_str(filehandle))))
            g.add((huid, P.tag, taguri))

        g.add((huid, P.qseqid, make_uri(qseqid)))
        g.add((huid, P.sseqid, make_uri(sseqid)))
        g.add((huid, P.pident, Literal(float(pident))))
        g.add((huid, P.length, Literal(int(length))))
        g.add((huid, P.mismatch, Literal(int(mismatch))))
        g.add((huid, P.gapopen, Literal(int(gapopen))))
        g.add((huid, P.qstart, Literal(int(qstart))))
        g.add((huid, P.qend, Literal(int(qend))))
        g.add((huid, P.sstart, Literal(int(sstart))))
        g.add((huid, P.send, Literal(int(send))))
        g.add((huid, P.evalue, Literal(float(evalue))))
        g.add((huid, P.bitscore, Literal(float(bitscore))))
    g.commit()


def mk_influenza_na(g, filehandle) -> None:
    def extract_strain(x):
        strain_pat = re.compile("[ABCD]/[^()\[\]]+")
        m = re.search(strain_pat, x)
        if m:
            return m.group(0)
        else:
            return None

    for line in tqdm(filehandle.readlines()):
        els = line.split("\t")
        try:
            classes.Phrase(
                [
                    flu.Genbank(els[0]),
                    tok.Unknown(els[1].lower(), field_name="host"),
                    flu.SegmentNumber(els[2]),
                    flu.Subtype(els[3]),
                    flu.Country(els[4]),
                    flu.Date(els[5]),
                    tok.Integer(els[6].lower(), field_name="length"),
                    flu.Strain(extract_strain(els[7])),
                    # skip 8
                    # skip 9
                    tok.Unknown(els[10].strip(), field_name="genome_status"),
                ]
            ).connect(g)
        except IndexError:
            log(line)
            sys.exit(1)


def mk_ird(g, filehandle) -> None:
    na_str = "-N/A-"
    for line in tqdm(filehandle.readlines()):
        els = line.split("\t")
        try:
            classes.Phrase(
                [
                    flu.SegmentNumber(els[0], na_str=na_str),
                    # skip protein name
                    flu.Genbank(els[2], field_name="genbank_id", na_str=na_str),
                    # skip complete genome
                    tok.Integer(els[4], field_name="length", na_str=na_str),
                    flu.Subtype(els[5], na_str=na_str),
                    flu.Date(els[6], na_str=na_str),
                    flu.Unknown(
                        els[7].replace("IRD:", "").lower(),
                        field_name="host",
                        na_str=na_str,
                    ),
                    flu.Country(els[8]),
                    # ignore state - can parse it from strain name
                    tok.Unknown(els[10], field_name="flu_season", na_str=na_str),
                    flu.Strain(els[11], field_name="strain_name", na_str=na_str),
                    # curation report - hard pass
                    ### I don't need your annotations, I can do my own, thank you very much
                    #  flu.US_Clade(els[13], field_name="us_clade", na_str=na_str),
                    #  flu.GlobalClade(els[14], field_name="gl_clade", na_str=na_str),
                ]
            ).connect(g)
        except IndexError:
            log(line)
            sys.exit(1)


def mk_gis(g, filehandle) -> None:
    if isinstance(filehandle, str):
        filename = filehandle
    else:
        filename = filehandle.name

    fh = pd.read_excel(filename, sheet_name=0)
    d = {c: [x for x in fh[c]] for c in fh}
    epipat = re.compile(" *\|.*")
    for i in tqdm(range(len(d["Isolate_Id"]))):
        try:
            epi_isl_id_tok = flu.Isolate(d["Isolate_Id"][i])

            # remove the parenthesized garbage following the strain name
            strain_clean = identifier.p_strain.parse(d["Isolate_Name"][i])
            # don't use Strain token here, to avoid double linking
            strain_tok = flu.Unknown(strain_clean, field_name="strain_name")
            # and keep the full strain name, even if ugly
            full_strain_name_tok = flu.Unknown(
                d["Isolate_Name"][i], field_name="gisaid_strain_name"
            )

            host_tok = flu.Host(d["Host"][i], field_name="host")
            subtype_tok = flu.Subtype(d["Subtype"][i], field_name="gisaid_subtype")
            lineage_tok = tok.Unknown(
                d["Lineage"][i], field_name="lineage", na_str=["", None]
            )
            try:
                country_tok = flu.Country(d["Location"][i].split(" / ")[1])
            except:
                country_tok = flu.Country(None)
            date_tok = flu.Date(d["Collection_Date"][i], field_name="collection_date")
            try:
                submission_date_tok = flu.Date(
                    d["Submission_Date"][i], field_name="submission_date"
                )
            except:
                submission_date_tok = flu.Date(None, field_name="submission_date")
            for segment in ("PB2", "PB1", "PA", "HA", "NP", "NA", "MP", "NS"):
                segment_tok = flu.SegmentName(segment)
                try:
                    epi_ids = [
                        re.sub(epipat, "", x)
                        for x in d[segment + " Segment_Id"][i].split(",")
                    ]
                except:
                    continue
                try:
                    gbk_ids = d[segment + " INSDC_Upload"][i].split(",")
                except:
                    gbk_ids = [None]
                for (epi_id, gbk_id) in zip(epi_ids, gbk_ids):
                    classes.Phrase(
                        [
                            epi_isl_id_tok,
                            flu.EpiSeqid(epi_id),
                            flu.Genbank(gbk_id),
                            strain_tok,
                            full_strain_name_tok,
                            segment_tok,
                            subtype_tok,
                            lineage_tok,
                            host_tok,
                            country_tok,
                            date_tok,
                            submission_date_tok,
                        ]
                    ).connect(g)
        except IndexError:
            log("Bad line - index error")
            for name, col in d.items():
                log(name + " : " + str(col[i]))
            sys.exit(1)
        except KeyError as e:
            log("This does not appear to be a valid gisaid metadata file")
            log(str(e))
            sys.exit(1)
        except:
            log("Bad line - other error")
            for name, col in d.items():
                log(name + " : " + str(col[i]))


def default_access(row, field):
    try:
        return row[field]["value"].split("+")
    except:
        return []


def append_add(entry, field, values):
    if len(values) > 0:
        if field in entry:
            for value in values:
                if not value in entry[field]:
                    entry[field].append(value)
        else:
            entry[field] = values
    elif field not in entry:
        entry[field] = []


def quarter_from_date(date):
    year, month = date.split("-")[0:2]
    quarter = str(math.ceil(int(month) / 3))
    return f"{year}Q{quarter}"


def _ustr(s):
    return s.upper().strip()


def _clean_subtype(s):
    try:
        (ha, na) = re.search(".*(H\\d+).*(N\\d+).*", _ustr(s)).groups()
        return ha + na
    except:
        return ""


def _get_subtype(strain, has, nas, gisaid_subtypes, genbank_subtypes):
    # subtype annotations that where either determined in the past or retrieved from epiflu
    gisaid_subtypes = list(
        {_clean_subtype(subtype) for subtype in gisaid_subtypes if len(subtype) > 0}
    )
    # genbank_subtypes annotations from genbank
    genbank_subtypes = list(
        {
            _clean_subtype(genbank_subtype)
            for genbank_subtype in genbank_subtypes
            if len(genbank_subtype) > 0
        }
    )
    # unique ha and na clades from octoFLU
    has = list({_ustr(ha) for ha in has if len(ha) > 0})
    nas = list({_ustr(na) for na in nas if len(na) > 0})

    # good subtype - trust octoFLU more than anything
    if len(nas) == 1 and len(has) == 1:
        subtype = has[0] + nas[0]
    # no subtype found based on HA/NA subtypes, so use genbank subtype field
    elif (
        len(gisaid_subtypes) > 1
        or len(genbank_subtypes) > 1
        or len(nas) > 1
        or len(has) > 1
    ):
        return "mixed"
    else:
        # if we have no octoFLU data
        # prefer genebank data over subtype data (e.g., from epiflu)
        if len(genbank_subtypes) == 1:
            subtype = genbank_subtypes[0]
        elif len(gisaid_subtypes) == 1:
            subtype = gisaid_subtypes[0]
        else:
            subtype = None

    return subtype


def mk_subtypes(results):
    entries = dict()

    for row in results["results"]["bindings"]:
        strain = row["strain_name"]["value"]

        if strain not in entries:
            entry = dict(
                isolates=set(),
                ha_subtypes=[],
                na_subtypes=[],
                subtypes=[],
                serotypes=[],
            )
        else:
            entry = entries[strain]

        isolates = entry["isolates"].update(
            [i for i in row["isolates"]["value"].split("+") if i]
        )

        append_add(entry, "genbank_subtypes", default_access(row, "genbank_subtypes"))
        append_add(entry, "gisaid_subtypes", default_access(row, "gisaid_subtypes"))

        maybe_segment_subtype = default_access(row, "segment_subtypes")
        if len(maybe_segment_subtype) == 1:
            segment_subtype = maybe_segment_subtype[0]
            if re.fullmatch("H\d+", segment_subtype):
                append_add(entry, "ha_subtypes", [segment_subtype])
            elif re.fullmatch("N\d+", segment_subtype):
                append_add(entry, "na_subtypes", [segment_subtype])

        entries[strain] = entry

    strain_entries = []
    isolate_entries = []
    for strain, entry in entries.items():
        subtype = _get_subtype(
            strain,
            has=entry["ha_subtypes"],
            nas=entry["na_subtypes"],
            gisaid_subtypes=entry["gisaid_subtypes"],
            genbank_subtypes=entry["genbank_subtypes"],
        )
        if subtype:
            strain_entries.append((strain, subtype))
            for isolate in entry["isolates"]:
                isolate_entries.append((isolate, subtype))

    return (strain_entries, isolate_entries)


MASTERLIST_HEADER = [
    "Barcode",
    "Date",
    "Collection_Q",
    "State",
    "Subtype",
    "H_Genbank",
    "N_Genbank",
    "PB2_Genbank",
    "PB1_Genbank",
    "PA_Genbank",
    "NP_Genbank",
    "M_Genbank",
    "NS_Genbank",
    "Strain",
    "US_Clade",
    "GL_Clade",
    "H1",
    "H3",
    "N1",
    "N2",
    "PB2",
    "PB1",
    "PA",
    "NP",
    "M",
    "NS",
    "Constellation",
    "Motif",
    "Sa_Motif",
    "Sb_Motif",
    "Ca1_Motif",
    "Ca2_Motif",
    "Cb_Motif",
]


def mk_masterlist(results):
    entries = dict()

    for row in results["results"]["bindings"]:
        barcode = row["barcode"]["value"]
        if barcode in entries:
            entry = entries[barcode]
        else:
            entry = dict()
            # initialize fields
            for field in MASTERLIST_HEADER:
                entry[field] = []

        gb = row["genbank_id"]["value"]
        segment = row["segment"]["value"]

        genbank_id = default_access(row, "genbank_id")[0]
        segment = default_access(row, "segment")[0]
        subtype = default_access(row, "subtypes")[0]
        # Use the earliest date since occasionally people resequence a sample
        # and incorrectly set the sequence date to the collection date (among
        # other issues)
        date = default_access(row, "earliest_date")[0]
        states = default_access(row, "states")
        strains = default_access(row, "strains")
        us_clades = default_access(row, "us_clades")
        gl_clades = default_access(row, "gl_clades")
        consts = default_access(row, "consts")
        h3_motifs = default_access(row, "h3_motifs")
        sa_motifs = default_access(row, "sa_motifs")
        sb_motifs = default_access(row, "sb_motifs")
        ca1_motifs = default_access(row, "ca1_motifs")
        ca2_motifs = default_access(row, "ca2_motifs")
        cb_motifs = default_access(row, "cb_motifs")

        append_add(entry, "Date", [date])
        append_add(entry, "Collection_Q", [quarter_from_date(date)])
        append_add(entry, "State", states)

        if segment == "HA":
            append_add(entry, "H_Genbank", [genbank_id])
            append_add(entry, "US_Clade", us_clades)
            append_add(entry, "GL_Clade", gl_clades)
        elif segment == "NA":
            append_add(entry, "N_Genbank", [genbank_id])
            append_add(entry, segment, us_clades)
        else:
            append_add(entry, segment + "_Genbank", [genbank_id])
            append_add(entry, segment, us_clades)

        maybe_segment_subtype = default_access(row, "segment_subtypes")
        if maybe_segment_subtype:
            segment_subtype = maybe_segment_subtype[0]

            if segment_subtype == "H1":
                append_add(entry, "H1", us_clades)
            elif segment_subtype == "H3":
                append_add(entry, "H3", us_clades)

            if segment_subtype == "N1":
                append_add(entry, "N1", us_clades)
            elif segment_subtype == "N2":
                append_add(entry, "N2", us_clades)

        append_add(entry, "Strain", strains)
        append_add(entry, "Subtype", [subtype])
        append_add(entry, "Constellation", consts)
        append_add(entry, "Motif", h3_motifs)
        append_add(entry, "Sa_Motif", sa_motifs)
        append_add(entry, "Sb_Motif", sb_motifs)
        append_add(entry, "Ca1_Motif", ca1_motifs)
        append_add(entry, "Ca2_Motif", ca2_motifs)
        append_add(entry, "Cb_Motif", cb_motifs)

        entries[barcode] = entry

    print("\t".join(MASTERLIST_HEADER))

    for barcode, entry in entries.items():
        entry["Barcode"] = [barcode]
        row = [",".join([f for f in entry[field] if f]) for field in MASTERLIST_HEADER]
        print("\t".join(row))


class IrregularStrain(flu.StrainToken):
    """
    Matches anything and treats it as a strain.

    This is useful when you want to force a field to refer to a strain, such as
    when you are loading unpublished data that use idiosyncratic identifiers.
    """

    typename = "strain_id"
    parser = parsec.regex(".+")


class IrregularFasta(classes.Ragged):
    """
    Load a FASTA file where the first field is treated as a strain identifier.
    """

    def cast(self, data):
        strain_ids = [IrregularStrain(d[0]) for d in data]
        phrases = super().cast([d[1:] for d in data])
        for i in range(len(strain_ids)):
            phrases[i].tokens.append(strain_ids[i])
        return phrases

    def connect(self, g):
        for phrase in self.data:
            for token in phrase.tokens:
                if token.group == "sequence":
                    g.add((token.as_uri(), P.tag, make_tag_uri("unpublished")))
        super().connect(g)


class IrregularSegment(flu.SegmentToken):
    """
    Matches anything and treats it as a segment.

    Also see IrregularStrain
    """

    typename = None
    parser = parsec.regex(".+")


class IrregularSegmentTable(classes.Table):
    """
    Load a table; where the kth field is treated as a segment identifier.
    """

    def cast(self, data):
        segment_ids = data[self.header[0]]
        del data[self.header[0]]
        phrases = super().cast(data)
        for i in range(len(segment_ids)):
            phrases[i].tokens.append(IrregularSegment(segment_ids[i]))
        return phrases



import click
import collections
import signal
import sys
import os
from octofludb.util import log, die
from octofludb.version import __version__


def open_graph():
    from rdflib import Graph
    from octofludb.nomenclature import manager

    return Graph(namespace_manager=manager)


def with_graph(f, *args, filename=None, outfile=sys.stdout, **kwargs):
    g = open_graph()

    if filename is not None:
        with open(filename, "r") as fh:
            f(g, fh, *args, **kwargs)
    else:
        f(g, *args, **kwargs)

    g.commit()  # just in case we missed anything
    log("Serializing to turtle format ... ", end="")
    turtles = g.serialize(format="turtle")
    log("done")
    for l in turtles.splitlines():
        try:
            print(l.decode("utf-8"), file=outfile)
        except:
            print(l, file=outfile)
    g.close()


def open_file(path):
    if path:
        filehandle = open(args["<filename>"], "r")
    else:
        filehandle = sys.stdin
    return filehandle


def make_na(na_str):
    if na_str:
        na_list = na_str.split(",")
        if isinstance(na_list, list):
            na_list = [None] + na_str
        else:
            na_list = [None, na_list]
    else:
        na_list = [None]
    return na_list


# Thanks to Максим Стукало from https://stackoverflow.com/questions/47972638
# for the solution to getting the subcommands to order non-alphabetically
class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands


CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

url_opt = click.option("--url", help="GraphDB URL", default="http://localhost:7200")

filename_arg = click.argument("filename", type=click.Path(exists=True))

repo_name_opt = click.option("--repo", help="Repository name", default="octofludb")

tag_arg_pos = click.argument("tag")

tag_arg_opt = click.option(
    "--tag", help="A tag to associate with each identifier", type=str
)

segment_key_opt = click.option(
    "--segment-key",
    help="Treat the first column as a segment identifier. This is necessary for irregular segment identifiers (such as sequence checksums), for genbank or epiflu IDs, not special actio is needed, since octofludb will automatically recognize them.",
    is_flag=True,
    default=False,
)

sparql_filename_pos = click.argument("sparql_filename", type=click.Path(exists=True))

all_the_turtles = click.argument(
    "turtle_filenames", type=click.Path(exists=True), nargs=-1
)

header_opt = click.option(
    "--header",
    is_flag=True,
    help="Include a header of column names indata returned from query",
)

fasta_opt = click.option(
    "--fasta",
    is_flag=True,
    help="Return query as a fasta file where last column is sequence",
)

delimiter_opt = click.option(
    "--delimiter", help="The delimiter between fields in the header", default="|"
)


@click.command(
    name="init",
)
@url_opt
@repo_name_opt
def init_cmd(url, repo):
    """
    Initialize an empty octofludb database
    """
    import pgraphdb as db
    import requests
    import shutil
    import octofludb.script as script

    config_file = os.path.join(
        os.path.dirname(__file__), "data", "octofludb-config.ttl"
    )
    try:
        db.make_repo(config=config_file, url=url)
    except requests.exceptions.ConnectionError:
        print(f"Could not connect to a GraphDB database at {url}", file=sys.stderr)
        sys.exit(1)

    octofludb_home = os.path.join(os.path.expanduser("~"), ".octofludb")

    if not os.path.exists(octofludb_home):
        try:
            print(
                f" - Creating local configuration folder at '{octofludb_home}'",
                file=sys.stderr,
            )
            os.mkdir(octofludb_home)
        except FileExistsError:
            print(
                f" Failed to create {octofludb_home}",
                file=sys.stderr,
            )
            sys.exit(1)

    script.initialize_config_file()


def upload_gisaid(config, url, repo):
    import octofludb.script as script
    import octofludb.recipes as recipe

    epiflu_metafiles = script.epiflu_meta_files(config)
    skipped_meta = 0
    if epiflu_metafiles:
        for epiflu_metafile in epiflu_metafiles:
            outfile = os.path.basename(epiflu_metafile) + ".ttl"
            if os.path.exists(outfile) and os.path.getsize(outfile) > 0:
                skipped_meta += 1
            else:
                with open(outfile, "w") as f:
                    with_graph(recipe.mk_gis, epiflu_metafile, outfile=f)
                upload([outfile], url=url, repo=repo)
    else:
        log("No epiflu metafiles found")
    if skipped_meta > 0:
        log(
            f"Skipped {str(skipped_meta)} epiflu meta files where existing non-empty turtle files were found in the build directory"
        )

    epiflu_fastafiles = script.epiflu_fasta_files(config)
    skipped_fasta = 0
    if epiflu_fastafiles:
        for infile in epiflu_fastafiles:
            outfile = os.path.basename(infile) + ".ttl"
            if os.path.exists(outfile) and os.path.getsize(outfile) > 0:
                skipped_fasta += 1
            else:
                with open(outfile, "w") as f:
                    prep_fasta(filename=infile, outfile=f)
                    upload([outfile], url=url, repo=repo)
    else:
        log("No epiflu fasta found")

    if skipped_fasta > 0:
        log(
            f"Skipped {str(skipped_fasta)} epiflu fasta files where existing non-empty turtle files were found in the build directory"
        )


def upload_classifications(config, url, repo):
    import octofludb.script as script
    import pgraphdb as db

    # octoflu classifications of unclassified swine
    # * retrieve unclassified strains
    unclassified_fasta = "unclassified-swine.fna"
    unclassified_classes = "unclassified-swine.txt"
    unclassified_turtle = "unclassified-swine.ttl"

    sparql_file = script.get_data_file("fetch-unclassified-swine.rq")
    with open(unclassified_fasta, "w") as fastaout:
        fmt_query_cmd(
            sparql_filename=sparql_file,
            header=False,
            fasta=True,
            url=url,
            repo=repo,
            outfile=fastaout,
        )

    with open(unclassified_classes, "w") as classout:
        # feed them into runOctoFLU
        classify_and_write(unclassified_fasta, outfile=classout)

    with open(unclassified_turtle, "w") as turtleout:
        # print the results
        prep_table(unclassified_classes, outfile=turtleout)

    upload([unclassified_turtle], url=url, repo=repo)

    # infer constellations
    constellation_table = "constellations.txt"
    constellation_turtles = "constellations.ttl"

    delete_constellations = script.get_data_file("delete-constellations.rq")
    db.update(sparql_file=delete_constellations, url=url, repo_name=repo)

    with open(constellation_table, "w") as constout:
        make_const(url=url, repo=repo, outfile=constout)

    with open(constellation_turtles, "w") as turtleout:
        prep_table(constellation_table, outfile=turtleout)

    upload([constellation_turtles], url=url, repo=repo)


def upload_subtypes(config, url, repo):
    import octofludb.script as script
    import pgraphdb as db

    # infer subtypes
    subtypes_table = "subtypes.txt"
    with open(subtypes_table, "w") as subtypesout:
        make_subtypes(url=url, repo=repo, outfile=subtypesout)

    genbank_subtypes = "subtypes-genbank.txt"
    epiflu_subtypes = "subtypes-epiflu.txt"
    with open(subtypes_table, "r") as subtypesin:
        with open(genbank_subtypes, "w") as gh:
            with open(epiflu_subtypes, "w") as eh:
                for (i, row) in enumerate(subtypesin.readlines()):
                    row = row.strip()
                    if i == 0:
                        print("strain_name\tsubtype", file=gh)
                        print("isolate_id\tsubtype", file=eh)
                    else:
                        if "EPI_ISL" in row:
                            print(row, file=eh)
                        else:
                            print(row, file=gh)

    gturtles = "subtypes-genbank.ttl"
    eturtles = "subtypes-epiflu.ttl"
    with open(gturtles, "w") as gturtleout:
        prep_table(genbank_subtypes, outfile=gturtleout)
    with open(eturtles, "w") as eturtleout:
        prep_table(epiflu_subtypes, outfile=eturtleout)
    upload([gturtles, eturtles], url=url, repo=repo)


def upload_motifs(config, url, repo):
    import octofludb.script as script

    # find h1 motifs
    h1_motif_table = script.findMotifs(
        os.path.join(os.path.dirname(__file__), "data", "get-h1-swine.rq"),
        [
            "sa_motif=124,125,155,157,159,160,162,163,164",
            "sb_motif=153,156,189,190,193,195",
            "ca1_motif=166,170,204,237",
            "ca2_motif=137,140,142,221,222",
            "cb_motif=70,71,73,74,75,115",
        ],
        "H1",
        url=url,
        repo_name=repo,
    )

    with open("h1-motifs.ttl", "w") as turtleout:
        prep_table(h1_motif_table, outfile=turtleout)
    upload(["h1-motifs.ttl"], url=url, repo=repo)

    # find h3 motifs
    h3_motif_table = script.findMotifs(
        os.path.join(os.path.dirname(__file__), "data", "get-h3-swine.rq"),
        ["h3_motif=145,155,156,158,189"],
        "H3",
        url=url,
        repo_name=repo,
    )

    with open("h3-motifs.ttl", "w") as turtleout:
        prep_table(h3_motif_table, outfile=turtleout)
    upload(["h3-motifs.ttl"], url=url, repo=repo)


@click.command(
    name="pull",
)
@click.option(
    "--nmonths",
    help="Update Genbank files for the last N months",
    default=1,
    type=click.IntRange(min=0, max=9999),
)
@click.option(
    "--no-schema", is_flag=True, default=False, help="Skip upload schema steps"
)
@click.option(
    "--no-clades",
    is_flag=True,
    default=False,
    help="Skip clade and constellation classification steps",
)
@click.option(
    "--no-subtype",
    is_flag=True,
    default=False,
    help="Skip subtype inferrence step",
)
@click.option(
    "--no-motifs", is_flag=True, default=False, help="Skip motif extraction steps"
)
@click.option(
    "--include-gisaid", is_flag=True, default=False, help="include gisaid data step"
)
@click.option(
    "--include-tags",
    is_flag=True,
    default=False,
    help="Upload tags as defined in the config file",
)
@url_opt
@repo_name_opt
def pull_cmd(
    nmonths,
    no_schema,
    no_clades,
    no_subtype,
    no_motifs,
    include_gisaid,
    include_tags,
    url,
    repo,
):
    """
    Update data. Pull from genbank, process any new data in the data folder,
    assign clades to swine data, assign subtypes to all data, and extract
    motifs.

    To build the database from nothing call `octofludb pull --nmonths=360`.
    This will pull all genbank data that has been released in the last 30
    years (which should be all of it).
    """
    import octofludb.recipes as recipe
    import octofludb.script as script
    import pgraphdb as db

    cwd = os.getcwd()

    script.gotoBuildHome()

    config = script.load_config_file()

    if not no_schema:
        # upload ontological schema
        schema_file = script.get_data_file("schema.ttl")  #
        upload([schema_file], url=url, repo=repo)

        # upload geological relationships
        geog_file = upload([script.get_data_file("geography.ttl")], url=url, repo=repo)[
            0
        ]

    if nmonths > 0:
        # update genbank (take a parameter telling how far back to go)
        # this command fills the current directory with .gb* files
        gb_turtles = prep_update_gb(minyear=1900, maxyear=2121, nmonths=nmonths)
        upload(gb_turtles, url=url, repo=repo)

    if include_gisaid:
        upload_gisaid(config, url, repo)

    if not no_clades:
        upload_classifications(config, url, repo)

    if not no_subtype:
        upload_subtypes(config, url, repo)

    if not no_motifs:
        upload_motifs(config, url, repo)

    if include_tags:
        # load all tags
        for (tag, basename) in config["tags"].items():
            for filename in script.tag_files(config, tag):
                outfile = filename + ".ttl"
                with open(outfile, "w") as f:
                    prep_tag(tag, filename, outfile=f)
                upload([outfile], url=url, repo=repo)

    os.chdir(cwd)

    return None


def fmt_query_cmd(sparql_filename, header, fasta, url, repo, outfile=sys.stdout):
    import octofludb.formatting as formatting
    import pgraphdb as db

    results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)
    if fasta:
        formatting.write_as_fasta(results, outfile=outfile)
    else:
        formatting.write_as_table(results, header=header, outfile=outfile)

    return outfile


@click.command(
    name="query",
)
@sparql_filename_pos
@header_opt
@fasta_opt
@url_opt
@repo_name_opt
def query_cmd(*args, **kwargs):
    """
    Submit a SPARQL query to octofludb
    """
    fmt_query_cmd(*args, **kwargs)


@click.command(name="classify")
@filename_arg
@click.option("--reference", help="An octoFLU reference fasta file", default=None)
def classify_cmd(filename, reference=None):
    """
    Classify the sequences in a fasta file using octoFLU

    The reference file used will be selected as follows:

      1. If --reference=REFERENCE is given, use REFERENCE. If the file does not exist, die.

      2. If the term `octoflu_references` is defined in the `config.yaml` file
         in octoflu home, then use this file. If term is not null and does not
         exist, die.

      3. If no references are given, use the default reference in the octoFLU repo.
    """
    classify_and_write(filename, reference=None, outfile=sys.stdout)


def classify_and_write(filename, reference=None, outfile=sys.stdout):
    import octofludb.colors as colors

    rows = classify(filename, reference=reference)
    print("seqid\tsegment_subtype\tclade\tgl_clade", file=outfile)

    # This may be empty, that is fine. A table with only a heade line would
    # generate an empty turtle file. Uploading an empty turtle file does
    # nothing.
    for row in rows:
        print("\t".join(row), file=outfile)


def classify(filename, reference=None):
    import octofludb.script as script

    if not reference:
        config = script.load_config_file()
        reference = script.get_octoflu_reference(config)
    return script.runOctoFLU(filename, reference)


@click.command(
    name="construct",
)
@sparql_filename_pos
@url_opt
@repo_name_opt
def construct_cmd(sparql_filename, url, repo):
    """
    Construct new triples
    """
    import pgraphdb as db

    results = db.sparql_construct(sparql_file=sparql_filename, url=url, repo_name=repo)

    print(results)

    return None


@click.command(
    name="update",
)
@sparql_filename_pos
@url_opt
@repo_name_opt
def update_cmd(sparql_filename, url, repo):
    """
    Submit a SPARQL delete or insert query to octofludb
    """
    import pgraphdb as db

    db.update(sparql_file=sparql_filename, url=url, repo_name=repo)
    return None


@click.command(
    name="upload",
)
@all_the_turtles
@url_opt
@repo_name_opt
def upload_cmd(turtle_filenames, url, repo):
    """
    Upload one or more turtle files to the database
    """
    upload(turtle_filenames, url, repo)


def upload(turtle_filenames, url, repo):
    import pgraphdb as db
    import octofludb.script as script

    files = []
    for filenames in turtle_filenames:
        for filename in script.expandpath(filenames):
            log(f"loading file: {filename}")
            db.load_data(url=url, repo_name=repo, turtle_file=filename)
            files.append(filename)
    return files


# ===== prep subcommands ====


@click.command(
    name="tag",
)
@click.argument("tag", type=str)
@filename_arg
def prep_tag_cmd(tag, filename):
    """
    Associate list of IDs with a tag
    """
    prep_tag(tag, filename)


def prep_tag(tag, filename, outfile=sys.stdout):
    from rdflib import Literal
    import datetime as datetime
    from octofludb.nomenclature import make_uri, make_tag_uri, P
    from octofludb.util import file_str

    g = open_graph()
    with open(filename, "r") as fh:
        taguri = make_tag_uri(tag)
        g = open_graph()
        g.add((taguri, P.name, Literal(tag)))
        g.add((taguri, P.time, Literal(datetime.datetime.now())))
        g.add((taguri, P.file, Literal(file_str(fh))))
        for identifier in (s.strip() for s in fh.readlines()):
            g.add((make_uri(identifier), P.tag, taguri))
        g.commit()  # just in case we missed anything
        log("Serializing to turtle format ... ", end="")
        turtles = g.serialize(format="turtle")
        log("done")
        for l in turtles.splitlines():
            try:
                print(l.decode("utf-8"), file=outfile)
            except:
                print(l, file=outfile)
    g.close()


@click.command(
    name="ivr",
)
@filename_arg
def prep_ivr_cmd(filename):
    """
    Translate an IVR table to RDF.

    load big table from IVR, with roughly the following format:
    gb | host | - | subtype | date | - | "Influenza A virus (<strain>(<subtype>))" | ...
    """
    import octofludb.recipes as recipe

    with_graph(recipe.mk_influenza_na, filename)


@click.command(
    name="ird",
)
@filename_arg
def prep_ird_cmd(filename):
    """
    Translate an IRD table to RDF.
    """

    import octofludb.recipes as recipe

    with_graph(recipe.mk_ird, filename)


@click.command(
    name="gis",
)
@filename_arg
def prep_gis_cmd(filename):
    """
    Translate a Gisaid metadata excel file to RDF.

    "filename" is a path to a Gisaid metadata excel file
    """
    import octofludb.recipes as recipe

    with_graph(recipe.mk_gis, filename)


def _mk_gbids_cmd(g, gbids=[]):
    import octofludb.entrez as entrez
    import octofludb.genbank as gb
    import octofludb.script as script

    error_msgs = []

    for gb_metas in entrez.get_gbs(gbids):
        for gb_meta in gb_metas:
            error_msg = gb.add_gb_meta_triples(g, gb_meta)
            if error_msg:
                error_msgs.append(error_msg)
        # commit the current batch (say of 1000 entries)
        g.commit()

    if len(error_msgs) > 0:
        logpath = script.error_log_entry(error_msgs, "failed_genbank_parses.txt")
        log(f"{len(error_msgs)} genbank entries could not be parsed, see {logpath}")


@click.command(
    name="gbids",
)
@filename_arg
def prep_gbids_cmd(*args, **kwargs):
    """
    Retrieve data for a list of genbank ids.

    <filename> contains a list of genbank ids
    """
    with open(filename, "r") as fh:
        gbids = [gbid.strip() for gbid in fh]
    log(f"Retrieving and parsing genbank ids from '{args.filename}'")
    with_graph(_mk_gbids_cmd, gbids=gbids)


@click.command(name="update_gb")
@click.option(
    "--minyear",
    help="Earliest year to update",
    default=1918,
    type=click.IntRange(
        min=1900, max=3021
    ),  # yes, octofludb will be used for a thousand years
)
@click.option(
    "--maxyear",
    help="Latest year to update",
    default=3021,
    type=click.IntRange(min=1900, max=3021),
)
@click.option(
    "--nmonths",
    help="Update Genbank files for the last N months",
    default=1440,
    type=click.IntRange(min=1, max=9999),
)
def prep_update_gb_cmd(minyear, maxyear, nmonths):
    """
    Retrieve any missing genbank records. Results are stored in files with the prefix '.gb_###.ttl'
    """
    return prep_update_gb(minyear, maxyear, nmonths)


def prep_update_gb(minyear, maxyear, nmonths):
    from octofludb.entrez import missing_acc_by_date
    import octofludb.colors as colors

    outfiles = []

    for date, missing_acc in missing_acc_by_date(
        min_year=minyear, max_year=maxyear, nmonths=nmonths
    ):
        if missing_acc:
            outfile = ".gb_" + date.replace("/", "-") + ".ttl"
            if os.path.exists(outfile) and os.path.getsize(outfile) > 0:
                log(f"GenBank turtle file for '{str(date)}' already exists, skipping")
            else:
                log(colors.good(f"Updating {date} ..."))
                with open(outfile, "w") as fh:
                    with_graph(_mk_gbids_cmd, gbids=missing_acc, outfile=fh)
                outfiles.append(outfile)
        else:
            log(colors.good(f"Up-to-date for {date}"))

    return outfiles


@click.command(
    name="blast",
)
@tag_arg_opt
@filename_arg
def prep_blast_cmd(tag, filename):
    """
    Translate BLAST results into RDF.

    <filename> File containing a list of genbank ids
    """
    import octofludb.recipes as recipe

    log(f"Retrieving and parsing blast results from '{filename}'")
    with_graph(recipe.mk_blast, filename, tag=tag)


def process_tablelike(include, exclude, levels):
    if not include:
        inc = {}
    else:
        inc = set(include.split(","))
    if not exclude:
        exc = {}
    else:
        exc = set(exclude.split(","))
    if not levels:
        levels = None
    else:
        levels = {s.strip() for s in levels.split(",")}
    return (inc, exc, levels)


include_opt = click.option(
    "--include", help="Only parse using these tokens (comma-delimited list)", default=""
)
exclude_opt = click.option(
    "--exclude", help="Remove these tokens (comma-delimited list)", default=""
)
na_opt = click.option("--na", help="The string that represents a missing value")


@click.command(
    name="table",
)
@filename_arg
@tag_arg_opt
@include_opt
@exclude_opt
@click.option("--levels", help="levels")
@na_opt
@segment_key_opt
def prep_table_cmd(*args, **kwargs):
    """
    Translate a table to RDF
    """
    prep_table(*args, **kwargs)


def prep_table(
    filename,
    tag=None,
    include=None,
    exclude=None,
    levels=None,
    na=None,
    segment_key=None,
    outfile=sys.stdout,
):
    """
    Translate a table to RDF
    """
    from octofludb.recipes import IrregularSegmentTable
    import octofludb.classes as classes

    if segment_key:
        constructor = IrregularSegmentTable
    else:
        constructor = classes.Table

    def _mk_table_cmd(g, fh):
        (inc, exc, levelsProc) = process_tablelike(include, exclude, levels)
        constructor(
            filehandle=fh,
            tag=tag,
            include=inc,
            exclude=exc,
            log=True,
            levels=levelsProc,
            na_str=make_na(na),
        ).connect(g)

    with_graph(_mk_table_cmd, filename=filename, outfile=outfile)


@click.command(
    name="fasta",
)
@filename_arg
@tag_arg_opt
@delimiter_opt
@include_opt
@exclude_opt
@na_opt
def prep_fasta_cmd(*args, **kwargs):
    """
    Translate a fasta file to RDF.

    <filename> Path to a TAB-delimited or excel table
    """

    return prep_fasta(*args, **kwargs)


def prep_fasta(
    filename,
    tag=None,
    delimiter=None,
    include=None,
    exclude=None,
    na=None,
    outfile=sys.stdout,
):
    import octofludb.classes as classes

    def _mk_fasta_cmd(g, fh):
        (inc, exc, levels) = process_tablelike(include, exclude, None)
        classes.Ragged(
            filehandle=fh,
            tag=tag,
            include=inc,
            exclude=exc,
            log=True,
            levels=levels,
            na_str=make_na(na),
        ).connect(g)

    with open(filename, "r") as fasta_fh:
        with_graph(_mk_fasta_cmd, fasta_fh, outfile=outfile)

    return outfile


@click.command(
    name="unpublished",
)
@filename_arg
@tag_arg_opt
@delimiter_opt
@include_opt
@exclude_opt
@na_opt
def prep_unpublished_cmd(filename, tag, delimiter, include, exclude, na):
    """
    Prepare an unpublished set up sequences.

    The input is a fasta file where the header is a series of terms separated
    by a delimiter ("|" by default).

    The first term MUST be the strain ID. This can be anything. For example,
    "A/swine/12345678/2020' or some arbitrary id.  If the ID is used elsewhere
    in the database to refer to a strain, then all data loaded here will be
    assumed to describe the other ID as well (i.e., they are considered to be
    the same thing).

    The sequence is assumed to be a segment of unknown subtype and clade. It
    will be associated with the strain by its MD5 checksum. Subtype/clade info
    can be added through octoFLU.

    Additional terms after the strain ID may be added. Any term that looks like
    a date (e.g., "2020-12-31") will be parsed as the collection date. Country
    names like "United States" or 3-letter country codes (e.g., USA or CAN) are
    supported.

    I strongly recommend you skim the output turtle file before uploading to
    the database.

    The "unpublished" tag is automatically associated with the segments in
    addition to any tag specified through the `--tag` option.
    """
    import octofludb.recipes as recipe

    def _mk_unpublished_fasta_cmd(g, fh):
        (inc, exc, levels) = process_tablelike(include, exclude, None)
        recipe.IrregularFasta(
            filehandle=fh,
            tag=tag,
            include=inc,
            exclude=exc,
            log=True,
            levels=levels,
            na_str=make_na(na),
        ).connect(g)

    with_graph(_mk_unpublished_fasta_cmd, filename=filename)


@click.group(
    cls=OrderedGroup,
    name="prep",
    context_settings=CONTEXT_SETTINGS,
)
def prep_grp():
    """
    Various recipes for prepping data for uploading.
    """
    pass


prep_grp.add_command(prep_update_gb_cmd)
prep_grp.add_command(prep_fasta_cmd)
prep_grp.add_command(prep_table_cmd)
prep_grp.add_command(prep_unpublished_cmd)
prep_grp.add_command(prep_tag_cmd)
prep_grp.add_command(prep_blast_cmd)
prep_grp.add_command(prep_gbids_cmd)
prep_grp.add_command(prep_gis_cmd)
prep_grp.add_command(prep_ird_cmd)
prep_grp.add_command(prep_ivr_cmd)


def make_const(url, repo, outfile=sys.stdout):
    """
    Generate constellations for all swine strains.

    A constellation is a succinct description of the internal 6 genes. The
    description consists of 6 symbols representing the phylogenetic clades of
    the 6 proteins: PB2, PB1, PA, NP, M, and NS. Current US strains should
    consist of genes from 3 groups: pandemic (P), TRIG (T), and the LAIV
    vaccine strain (V). "-" indicates that no sequence is available for the
    given segment. "H" represents a human seasonal internal gene (there are
    only a few of these in the US. "X" represents something else exotic (e.g.,
    not US). For mixed strains, the constellation will be recorded as "mixed".
    """
    import octofludb.formatting as formatting
    import pgraphdb as db

    sparql_filename = os.path.join(os.path.dirname(__file__), "data", "segments.rq")
    results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)
    formatting.write_constellations(results, outfile=outfile)


def make_subtypes(url, repo, outfile=sys.stdout):
    strains, isolates = get_missing_subtypes(url, repo)
    print("strain_name\tsubtype", file=outfile)
    for identifier, subtype in strains + isolates:
        print(f"{identifier}\t{subtype}", file=outfile)


def get_missing_subtypes(url, repo):
    import octofludb.recipes as recipe
    import pgraphdb as db

    sparql_filename = os.path.join(os.path.dirname(__file__), "data", "subtypes.rq")

    results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)

    return recipe.mk_subtypes(results)


@click.command(
    name="masterlist",
)
@url_opt
@repo_name_opt
def report_masterlist_cmd(url, repo):
    """
    Generate the surveillance masterlist
    """
    import octofludb.recipes as recipe
    import pgraphdb as db

    sparql_filename = os.path.join(os.path.dirname(__file__), "data", "masterlist.rq")

    results = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)

    recipe.mk_masterlist(results)


# ===== fetch subcommands =====


@click.command(
    name="tag",
)
@filename_arg
@url_opt
@repo_name_opt
def fetch_tag_cmd(filename, url, repo):
    """
    Upload list of tags
    """
    from rdflib import Literal
    from octofludb.nomenclature import make_query_tag_uri, P
    from tempfile import mkstemp

    g = open_graph()
    # get all identifiers
    with open(filename, "r") as fh:
        identifiers = [s.strip() for s in fh.readlines()]

    # make the turtle file
    taguri = make_query_tag_uri()
    g = open_graph()
    for identifier in identifiers:
        g.add((taguri, P.query_tag, Literal(identifier)))
    g.commit()  # just in case we missed anything
    log("Serializing to turtle format ... ", end="")
    turtles = g.serialize(format="turtle")
    log("done")
    (n, turtle_filename) = mkstemp(suffix=".ttl")
    with open(turtle_filename, "w") as th:
        for l in turtles.splitlines():
            try:
                print(l.decode("utf-8"), file=th)
            except:
                print(l, file=th)

    # upload it to the database
    upload_cmd([turtle_filename], url, repo)
    g.close()

    return None


@click.command(
    name="isolate",
)
@url_opt
@repo_name_opt
def fetch_isolate_cmd(url, repo):
    """
    Fetch tagged isolate data
    """
    sparql_filename = os.path.join(
        os.path.dirname(__file__), "data", "get-tagged-isolate.rq"
    )
    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)


@click.command(
    name="strain",
)
@url_opt
@repo_name_opt
def fetch_strain_cmd(url, repo):
    """
    Fetch tagged strain data
    """
    sparql_filename = os.path.join(
        os.path.dirname(__file__), "data", "get-tagged-strain.rq"
    )
    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)


@click.command(
    name="segment",
)
@url_opt
@repo_name_opt
def fetch_segment_cmd(url, repo):
    """
    Fetch tagged segment data
    """
    sparql_filename = os.path.join(
        os.path.dirname(__file__), "data", "get-tagged-segment.rq"
    )
    fmt_query_cmd(sparql_filename, header=True, fasta=False, url=url, repo=repo)


@click.command(
    name="sequence",
)
@url_opt
@repo_name_opt
def fetch_sequence_cmd(url, repo):
    """
    Fetch tagged sequence data
    """
    sparql_filename = os.path.join(
        os.path.dirname(__file__), "data", "get-tagged-sequence.rq"
    )
    fmt_query_cmd(sparql_filename, header=False, fasta=True, url=url, repo=repo)


@click.command(
    name="clear",
)
@url_opt
@repo_name_opt
def fetch_clear_cmd(url, repo):
    """
    Clear all uploaded tags
    """
    import pgraphdb as db

    sparql_filename = os.path.join(
        os.path.dirname(__file__), "data", "clear-query-tags.rq"
    )
    db.update(sparql_file=sparql_filename, url=url, repo_name=repo)


@click.group(
    cls=OrderedGroup,
    name="fetch",
    context_settings=CONTEXT_SETTINGS,
)
def fetch_grp():
    """
    Tag and fetch data

    Use the "tag" subcommand to push lists of identifiers (as either a
    comma-delimited string or a file with one identifier per line).

    These identifiers will all be tagged for retrieval in the database.

    They may be fetched with the "isolate", "strain" or "segment" commands
    (dependeing on what is tagged).

    Tags can be cleared with `clear`. Until cleared, the tags reside in the
    database, allowing multiple retrievals or allowing other database
    operations to interact with the tagged sets (not a recommended operation).
    """
    pass


fetch_grp.add_command(fetch_tag_cmd)
fetch_grp.add_command(fetch_isolate_cmd)
fetch_grp.add_command(fetch_strain_cmd)
fetch_grp.add_command(fetch_segment_cmd)
fetch_grp.add_command(fetch_sequence_cmd)
fetch_grp.add_command(fetch_clear_cmd)


# ===== report subcommands =====


def macro_query(filename, macros, *args, **kwargs):
    import re
    from tempfile import mkstemp

    sparql_filename = os.path.join(os.path.dirname(__file__), "data", filename)

    (n, tmpfile) = mkstemp(suffix=".rq")

    with open(sparql_filename, "r") as template:
        with open(tmpfile, "w") as query:
            for line in template.readlines():
                for (macro, replacement) in macros:
                    line = re.sub(macro, replacement, line)
                print(line, file=query)

    fmt_query_cmd(tmpfile, *args, **kwargs)

    os.remove(tmpfile)


@click.command(
    name="monthly",
)
@click.argument("year", type=click.IntRange(min=2009, max=2100))
@click.argument("month", type=click.IntRange(min=1, max=12))
@click.option("--context", is_flag=True, help="Get contextualizing sequences")
@url_opt
@repo_name_opt
def report_monthly_cmd(year, month, context, url, repo):
    """
    Surveillance data for the given month (basis of WGS selections)
    """

    def pad(x):
        if x < 10 and x >= 0:
            return "0" + str(x)
        else:
            return str(x)

    if context:

        macros = [
            ("__MIN_DATE__", f"{str(year - 1)}-{pad(month)}-01"),
            ("__MAX_DATE__", f"{str(year + 1)}-{pad(month)}-01"),
        ]

        macro_query(
            "monthly-context.rq", macros, header=True, fasta=True, url=url, repo=repo
        )

    else:

        macros = [("__YEAR__", str(year)), ("__MONTH__", str(month))]

        macro_query("wgs.rq", macros, header=True, fasta=False, url=url, repo=repo)


@click.command(
    name="quarter",
)
@url_opt
@repo_name_opt
def report_quarter_cmd(url, repo):
    """
    Surveillance data for the quarter (basis of quarterly reports)

    Currently, this just generates the smae masterlist as is used in
    octoflushow. However, it may eventually be specialized.
    """
    report_masterlist_cmd(url=url, repo=repo)


@click.command(
    name="offlu",
)
@url_opt
@repo_name_opt
def report_offlu_cmd(url, repo):
    """
    Synthesize public and private data needed for offlu reports
    """
    raise NotImplemented


@click.group(
    cls=OrderedGroup,
    name="report",
    context_settings=CONTEXT_SETTINGS,
)
def report_grp():
    """
    Build standardized reports from the data
    """
    pass


report_grp.add_command(report_offlu_cmd)
report_grp.add_command(report_monthly_cmd)
report_grp.add_command(report_quarter_cmd)
report_grp.add_command(report_masterlist_cmd)

# ===== deletion subcommands =====


@click.command(
    name="constellations",
)
@url_opt
@repo_name_opt
def delete_constellations_cmd(url, repo):
    """
    Delete all constellation data
    """
    import octofludb.script as script
    import pgraphdb as db

    delete_script = script.get_data_file("delete-constellations.rq")
    db.update(sparql_file=delete_script, url=url, repo_name=repo)


@click.command(
    name="subtypes",
)
@url_opt
@repo_name_opt
def delete_subtypes_cmd(url, repo):
    """
    Delete all subtype data
    """
    import octofludb.script as script
    import pgraphdb as db

    delete_script = script.get_data_file("delete-subtypes.rq")
    db.update(sparql_file=delete_script, url=url, repo_name=repo)


@click.command(
    name="us-clades",
)
@url_opt
@repo_name_opt
def delete_us_clades_cmd(url, repo):
    """
    Delete all clade data
    """
    import octofludb.script as script
    import pgraphdb as db

    delete_script = script.get_data_file("delete-us_clades.rq")
    db.update(sparql_file=delete_script, url=url, repo_name=repo)


@click.command(
    name="gl-clades",
)
@url_opt
@repo_name_opt
def delete_gl_clades_cmd(url, repo):
    """
    Delete all global H1 clade data
    """
    import octofludb.script as script
    import pgraphdb as db

    delete_script = script.get_data_file("delete-gl_clades.rq")
    db.update(sparql_file=delete_script, url=url, repo_name=repo)


@click.command(
    name="motifs",
)
@url_opt
@repo_name_opt
def delete_motifs_cmd(url, repo):
    """
    Delete all antigenic motifs
    """
    import octofludb.script as script
    import pgraphdb as db

    delete_script = script.get_data_file("delete-motifs.rq")
    db.update(sparql_file=delete_script, url=url, repo_name=repo)


@click.group(
    cls=OrderedGroup,
    name="delete",
    context_settings=CONTEXT_SETTINGS,
)
def delete_grp():
    """
    Delete datasets. This can be useful when the dataset is corrupted in some
    way. For example, if you changed clade names in your octoFLU reference file
    and then reran `octofludb pull`, no change would occur since only strains
    with missing clades are updated. Instead you would need to first delete the
    existing clades and restart. If instead you manually ran octoFLU, built a
    turtle file from the resulting file, and uploaded that, then you could end
    up with multiple clades associated with some sequences.
    """
    pass


delete_grp.add_command(delete_constellations_cmd)
delete_grp.add_command(delete_subtypes_cmd)
delete_grp.add_command(delete_us_clades_cmd)
delete_grp.add_command(delete_gl_clades_cmd)
delete_grp.add_command(delete_motifs_cmd)


@click.group(cls=OrderedGroup, context_settings=CONTEXT_SETTINGS)
@click.version_option(__version__, "-v", "--version", message=__version__)
def cli_grp():
    """
    API and utilities for the USDA swine IVA surveillance database
    """
    pass


cli_grp.add_command(init_cmd)
cli_grp.add_command(pull_cmd)
cli_grp.add_command(query_cmd)
cli_grp.add_command(update_cmd)
cli_grp.add_command(classify_cmd)
cli_grp.add_command(construct_cmd)
cli_grp.add_command(upload_cmd)
cli_grp.add_command(prep_grp)
cli_grp.add_command(fetch_grp)
cli_grp.add_command(report_grp)
cli_grp.add_command(delete_grp)


def main():
    cli_grp()


if __name__ == "__main__":
    if os.name is "posix":
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    main()



import itertools
import rdflib
import urllib.parse as url
import sys
import re
import octofludb.domain.geography as geo
import octofludb.domain.date as date
from rdflib.namespace import RDF, RDFS, OWL, XSD
from octofludb.util import padDigit

ni = rdflib.Namespace("https://flu-crew.org/id/")
nt = rdflib.Namespace("https://flu-crew.org/term/")
ntag = rdflib.Namespace("https://flu-crew.org/tag/")
nquery = rdflib.Namespace("https://flu-crew.org/query/")
nusa = rdflib.Namespace("https://flu-crew.org/geo/country/usa/")
ncountry = rdflib.Namespace("https://flu-crew.org/geo/country/")

manager = rdflib.namespace.NamespaceManager(rdflib.Graph())
manager.bind("fid", ni)
manager.bind("f", nt)
manager.bind("usa", nusa)
manager.bind("world", ncountry)
manager.bind("query", nquery)


def make_tag_uri(x):
    tag = x.strip().replace(" ", "_").lower()
    tag = url.quote_plus(tag)
    return ntag.term(tag)


def make_query_tag_uri(x="default"):
    tag = url.quote_plus(x)
    return nquery.term(tag)


def define_subproperty(p1, p2, g):
    """
    define p1 as a subproperty of p2 in graph g
    """
    if p1 != p2:
        g.add((p1, RDFS.subPropertyOf, p2))


def uidgen(base="_", pad=3, start=0):
    base = base.replace(" ", "_")
    for i in itertools.count(0):
        yield ni.term(padDigit(base + str(i), pad))


def make_uri(x, namespace=ni):
    if not x:
        return None
    if isinstance(x, rdflib.term.URIRef):
        return x
    else:
        x = re.sub("[ -]+", "_", x.strip()).lower()
        return namespace.term(url.quote_plus(x))


def make_usa_state_uri(code):
    abbr = geo.state_to_code(code)
    if not abbr:
        print(
            f"Expected a USA state name or postal abbreviation, found '{code}'",
            file=sys.stderr,
        )
        sys.exit(1)
    return nusa.term(abbr)


def make_country_uri(countryStr):
    code = geo.country_to_code(countryStr)
    if code:
        uri = ncountry.term(code)
    else:
        uri = make_uri(countryStr, namespace=ncountry)
    return uri


def make_country_uri_from_code(code):
    return ncountry.term(code)


def make_date(dateStr):
    try:
        # Parse this to a date if it is of the pandas date type
        # This will remove any time annotation
        dateStr = str(dateStr.date())
    except AttributeError:
        pass
    try:
        uri = date.p_any_date.parse_strict(dateStr).as_uri()
    except:
        uri = None
    return uri


def make_property(x):
    return nt.term(x.lower().replace(" ", "_"))


def make_literal(x, infer=True):
    if not infer:
        return rdflib.Literal(x)
    try:
        # Can x be a date?
        return rdflib.Literal(str(date.p_date.parse(x)), datatype=XSD.date)
    except:
        return rdflib.Literal(x)


class O:
    feature = nt.Feature
    unknown_strain = nt.unknown_strain
    unknown_unknown = nt.unknown


class P:
    # standard semantic web predicates
    name = nt.name  # in scheme: rdfs:label rdfs:subPropertyOf f:name
    abbr = nt.abbr
    sameAs = OWL.sameAs
    unknown_unknown = nt.unknown
    chksum = nt.chksum
    # flu relations
    has_feature = nt.has_feature
    tag = nt.tag
    query_tag = nt.query_tag
    dnaseq = nt.dnaseq
    proseq = nt.proseq
    global_clade = nt.global_clade
    constellation = nt.constellation
    segment_name = nt.segment_name
    segment_number = nt.segment_number
    unknown_strain = nt.unknown_strain
    # blast predicates
    qseqid = nt.qseqid
    sseqid = nt.sseqid
    pident = nt.pident
    length = nt.length
    mismatch = nt.mismatch
    gapopen = nt.gapopen
    qstart = nt.qstart
    qend = nt.qend
    sstart = nt.sstart
    send = nt.send
    evalue = nt.evalue
    bitscore = nt.bitscore
    # labels for sequences
    gb = nt.genbank_id
    epi_id = nt.epi_id
    # labels for strains
    strain_name = nt.strain_name
    barcode = nt.barcode
    epi_isolate = nt.epi_isolate
    has_segment = nt.has_segment
    # the local curated data
    ref_reason = nt.ref_reason
    country = nt.country
    country_name = nt.country_name
    state = nt.state
    subtype = nt.subtype
    ha_clade = nt.ha_clade
    na_clade = nt.na_clade
    date = nt.date
    time = nt.time
    file = nt.file
    host = nt.host
    encodes = nt.gene
    # -----------------------------------------------------------------------
    # gb/*  -- I need to start generalizing away from this, since this data
    # does not come only from genebank.
    # -----------------------------------------------------------------------
    feature_key = nt.feature_key
    gb_locus = nt.locus  # unique key
    gb_length = nt.length
    gb_strandedness = nt.strandedness
    gb_moltype = nt.moltype
    gb_topology = nt.topology
    gb_division = nt.division
    gb_update_date = nt.update_date
    gb_create_date = nt.create_date
    gb_definition = nt.definition
    gb_primary_accession = nt.primary_accession
    gb_accession_version = nt.accession_version
    gb_other_seqids = nt.other_seqids
    gb_source = nt.source
    gb_organism = nt.organism
    gb_taxonomy = nt.taxonomy
    gb_references = nt.references
    gb_sequence = nt.sequence
    # -----------------------------------------------------------------------
    # gb/feature/*
    # -----------------------------------------------------------------------
    # a set of features associated with this particular strain
    gb_key = nt.key  # feature type (source | gene | CDS | misc_feature)
    gb_location = nt.location
    gb_intervals = nt.intervals
    gb_operator = nt.operator
    # a set of qualifiers for this feature
    # in biopython, this is a list of
    # {'GBQualifier_value' 'GBQualifier_name'} dicts
    gb_codon_start = nt.codon_start
    gb_collection_date = nt.collection_date
    gb_country = nt.country
    gb_db_xref = nt.db_xref
    gb_gene = nt.gene
    gb_host = nt.host
    gb_isolation_source = nt.isolation_source
    gb_mol_type = nt.mol_type
    gb_note = nt.note
    gb_organism = nt.organism
    gb_product = nt.product
    gb_protein_id = nt.protein_id
    gb_serotype = nt.serotype
    gb_strain = nt.strain_name
    gb_transl_table = nt.transl_table
    gb_translation = nt.translation



from octofludb.util import log
from octofludb.colors import bad

import sys


def write_as_fasta(results, outfile=sys.stdout):
    """
    Write a SPARQL query result as a FASTA file
    """
    header_fields = results["head"]["vars"][:-1]
    seq_field = results["head"]["vars"][-1]
    for row in results["results"]["bindings"]:
        fields = []
        for f in header_fields:
            if f in row:
                fields.append(row[f]["value"])
            else:
                fields.append("")
        header = "|".join(fields)
        sequence = row[seq_field]["value"]
        print(">" + header, file=outfile)
        print(sequence, file=outfile)


def write_as_table(results, header=True, outfile=sys.stdout):
    """
    Write a SPARQL query result as a TAB-delimited table with an optional header
    """

    def val(xs, field):
        if field in xs:
            return xs[field]["value"]
        else:
            return ""

    if header:
        print("\t".join(results["head"]["vars"]), file=outfile)
    for row in results["results"]["bindings"]:
        fields = (val(row, field) for field in results["head"]["vars"])
        print("\t".join(fields), file=outfile)


def write_constellations(results, outfile=sys.stdout):
    """
    Prepare constellations
    """

    rows = _parse_constellation_query(results)

    consts = _make_constellations(rows)

    print("strain_name\tconstellation", file=outfile)
    for (strain, const) in consts:
        print(f"{strain}\t{const}", file=outfile)


def _parse_constellation_query(results):
    return [
        (row["strain"]["value"], row["segment"]["value"], row["clade"]["value"])
        for row in results["results"]["bindings"]
    ]


def _make_constellations(rows):

    segment_lookup = dict(PB2=0, PB1=1, PA=2, NP=3, M=4, MP=4, NS=5)

    clade_lookup = dict(
        pdm="P", LAIV="V", TRIG="T", humanSeasonal="H", classicalSwine="C"
    )

    const = dict()
    for (strain, segment, clade) in rows:

        if strain in const:
            if const[strain] is None:
                continue
        else:
            const[strain] = list("------")

        try:
            index = segment_lookup[segment]
        except KeyError as e:
            log(
                f"{bad('WARNING:')} segment/segment_subtype mismatch, {str((strain, segment, clade))}"
            )
            continue

        if clade in clade_lookup:
            char = clade_lookup[clade]
        else:
            char = "X"

        if const[strain][index] == "-":
            const[strain][index] = char
        elif const[strain][index] != char:
            const[strain] = None

    rows = []
    for (k, c) in const.items():
        if c is None:
            rows.append((k, "mixed"))
        else:
            rows.append((k, "".join(c)))
    return rows



from setuptools import setup

from octofludb.version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="octofludb",
    version=__version__,
    description="Mangage the flu-crew swine surveillance database",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flu-crew/octofludb",
    author="Zebulun Arendsee",
    author_email="zebulun.arendsee@usda.gov",
    packages=["octofludb", "octofludb.classifiers", "octofludb.domain"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": ["octofludb=octofludb.ui:main"]},
    install_requires=[
        "pgraphdb",
        "requests",
        "parsec",
        "rdflib",
        "pandas",
        "biopython",
        "tqdm",
        "xlrd",
        "flutile",
    ],
    py_modules=["octofludb"],
    zip_safe=False,
    include_package_data=True,
)



from typing import List
from rdflib import Literal
import math
import sys


def log(msg, end="\n"):
    print(msg, file=sys.stderr, flush=True, end=end)


def die(msg):
    print(msg, file=sys.stderr, flush=True)
    sys.exit(1)


def file_str(f):
    if isinstance(f, str):
        return f
    else:
        try:
            return f.name  # get the name from a filehandle
        except:
            return str(f)


def zipGen(x, y):
    if x is None:
        x = []
    if y is None:
        y = []
    x += [None] * (len(y) - len(x))
    y += [None] * (len(x) - len(y))
    return zip(x, y)


def strOrNone(x):
    try:
        if math.isnan(x):
            x = None
    except:
        pass
    if x is not None:
        x = str(x)
    return x


def identity(x):
    x


def upper(x):
    return x.upper()


def lower(x):
    return x.lower()


def underscore(x):
    return x.replace(" ", "_")


def strip(x):
    return x.strip()


def compose(*args):
    def composition(x):
        for f in reversed(args):
            x = f(x)
        return x

    return composition


def make_maybe_add(g, meta, sid):
    def maybe_add(p, key, formatter=Literal):
        if key in meta and meta[key] != None:
            try:
                g.add((sid, p, formatter(meta[key])))
            except:
                pass

    return maybe_add


def rmNone(xs):
    """Remove all 'None' elements from a list"""
    return list(filter(lambda x: x != None, xs))


def firstOne(xs):
    """Return the first defined value in a list"""
    return rmNone(xs)[0]


def concat(xs: List[str]) -> str:
    return "".join(xs)


def padDigit(x: str, n=2) -> str:
    """This is used, for example, to exapand the month '5' to '05'"""
    return "0" * (n - len(x)) + x


def replace(d, key, a, b):
    if d[key] != None:
        d[key] = d[key].replace(a, b)
    return d


def fixRegexMap(d: dict, field: str, rexpr: str, m: dict, flags=0):
    if d[field] != None:
        key = rmNone([re.fullmatch(rexpr, k, flags) for k in m.keys()])
        if len(key) > 0:
            d[field] = m[key[0].string]
    return d


def fixLookup(d: dict, field: str, m: dict, f=lambda x: x):
    try:
        d[field] = m[f(field)]
    except:
        pass
    return d


def addDefault(d: dict, key: str, default: str):
    if d[key] == None:
        d[key] = default
    return d



import sys
import parsec
from octofludb.classifiers.flucrew import allClassifiers
from octofludb.token import Token, Unknown, Missing
from octofludb.util import zipGen, strOrNone, log, concat, file_str
from octofludb.nomenclature import make_tag_uri, P
import dateutil.parser as dateparser
import xlrd
import pandas as pd
import octofludb.colors as colors
import datetime as datetime
from rdflib import Literal
from tqdm import tqdm


def updateClassifiers(classifiers, include, exclude):
    keys = list(classifiers.keys())
    for classifier in keys:
        if classifier in exclude:
            classifiers.pop(classifier)
        if classifier in include:
            classifiers.pop(classifier)
    return list(classifiers.values())


class Interpreter:
    def __init__(
        self,
        data,
        field_name=None,
        tag=None,
        classifiers=allClassifiers,
        default_classifier=Unknown,
        include={},
        exclude={},
        levels=None,
        na_str=[None],
        log=False,
    ):
        self.tag = tag
        self.levels = levels
        self.na_str = na_str
        self.classifiers = updateClassifiers(classifiers, include, exclude)
        self.default_classifier = Unknown
        if log:
            self.log()
        self.field_name = field_name
        self.data = self.cast(data)

    def cast(self, data):
        raise NotImplementedError

    def load(self, g):
        raise NotImplementedError

    def summarize(self):
        raise NotImplementedError

    def log(self):
        log("Parsing with the following tokens:")
        for classifier in self.classifiers:
            log(f"  {colors.good(classifier.typename)}")
        if self.tag:
            log(f"Tagging as '{self.tag}'")
        else:
            log(f"{colors.bad('No tag given')}")


class Datum(Interpreter):
    """
    Interpret a single word. This should not be used much.
    """

    def cast(self, data):
        if data == "":
            return Missing(data, na_str=self.na_str)
        for classifier in self.classifiers:
            try:
                token = classifier(data, field_name=self.field_name, na_str=self.na_str)
            except TypeError:
                log(data)
                log(token)
                sys.exit(1)
            if token:
                return token
        return self.default_classifier(
            data, field_name=self.field_name, na_str=self.na_str
        )

    def summarize(self):
        log(f"typename: {self.data.typename}")
        log(f"field_name: {self.data.field_name}")
        log(f"value: {self.data.dirty}")
        log(f"munged: {self.data.clean}")

    def __str__(self):
        return str(self.data.clean)


def addTag(g, tag, filehandle):
    """
    Add tag info to the triple set and return the tag URI
    """
    if tag:
        taguri = make_tag_uri(tag)
        g.add((taguri, P.name, Literal(tag)))
        g.add((taguri, P.time, Literal(datetime.datetime.now())))
        g.add((taguri, P.file, Literal(file_str(filehandle))))
    else:
        taguri = None
    return taguri


class HomoList(Interpreter):
    """
    Interpret a list of items assumed to be of the same type
    """

    def cast(self, data):
        for classifier in self.classifiers:
            if classifier.goodness(data, na_str=self.na_str) > 0.8:
                c = classifier
                break
        else:
            c = self.default_classifier
        return [c(x, field_name=self.field_name, na_str=self.na_str) for x in data]

    def connect(self, g):
        addTag(g, tag=self.tag, filehandle=filehandle)
        for token in self.data:
            if token.clean is None:
                continue
            token.add_triples(g)

    def __str__(self):
        return str([t.clean for t in self.data])


class ParsedPhraseList(Interpreter):
    def __init__(
        self,
        filehandle,
        field_name=None,
        tag=None,
        classifiers=allClassifiers,
        default_classifier=Unknown,
        include={},
        exclude={},
        levels=None,
        na_str=[None],
        log=False,
    ):
        self.classifiers = updateClassifiers(classifiers, include, exclude)
        self.tag = tag
        self.levels = levels
        self.na_str = na_str
        if log:
            self.log()
        self.filehandle = filehandle
        self.default_classifier = default_classifier
        self.data = self.cast(self.parse(filehandle))

    def parse(self, filehandle):
        raise NotImplementedError

    def connect(self, g):
        log("Making triples")
        taguri = addTag(g, tag=self.tag, filehandle=self.filehandle)
        for (i, phrase) in enumerate(tqdm(self.data)):
            phrase.connect(g, taguri=taguri)


def tabularTyping(data, levels=None, na_str=[None]):
    cols = []
    if not data:
        return []
    for k, v in data.items():
        hl = HomoList(v, field_name=k, na_str=na_str).data
        if len(hl) > 0:
            log(f" - '{k}':{colors.good(hl[0].typename)}")
        else:
            log(f"{colors.bad('Warning:')} no data")
        cols.append(hl)
    phrases = [
        Phrase([col[i] for col in cols], levels=levels) for i in range(len(cols[0]))
    ]
    return phrases


def headlessTabularTyping(data, levels=None, na_str=[None]):
    cols = []
    if not data:
        return []
    for (i, xs) in enumerate(data):
        hl = HomoList(xs, na_str=na_str).data
        log(f" - 'X{i}':{colors.good(hl[0].typename)}")
        cols.append(hl)
    phrases = [
        Phrase([col[i] for col in cols], levels=levels) for i in range(len(cols[0]))
    ]
    return phrases


class Table(ParsedPhraseList):
    """
    Parse a table.

    The table may be a TAB-delimited file or an excel file. It is assumed to
    have a header.
    """

    def cast(self, data):
        return tabularTyping(data, levels=self.levels, na_str=self.na_str)

    def parse(self, filehandle):
        """
        Make a dictionary with column name as key and list of strings as value.
        Currently only Excel is supported.
        """
        try:
            data = self._parse_excel(filehandle)
        except:
            data = self._parse_table(filehandle)
        return data

    def _parse_excel(self, filehandle):
        try:
            log(f"Reading {file_str(filehandle)} as excel file ...")
            d = pd.read_excel(filehandle.name)
            self.header = list(d.columns)
            # create a dictionary of List(str) with column names as keys
            return {c: [strOrNone(x) for x in d[c]] for c in d}
        except xlrd.biffh.XLRDError as e:
            log(f"Could not parse '{file_str(filehandle)}' as an excel file")
            raise e
        return d

    def _parse_table(self, filehandle, delimiter="\t"):
        log(f"Reading {file_str(filehandle)} as tab-delimited file ...")
        rows = [r.split(delimiter) for r in filehandle.readlines()]
        if len(rows) < 2:
            # if the input table is empty or has only a header
            return dict()
        else:
            self.header = [c.strip() for c in rows[0]]
            indices = range(len(self.header))
            rows = rows[1:]
            columns = {
                self.header[i]: [strOrNone(r[i].strip()) for r in rows] for i in indices
            }
            return columns


class Ragged(ParsedPhraseList):
    """
    Interpret a ragged list of lists (e.g. a fasta file). For now I will parse
    each sublist as a Phrase. I could probable extract some type information
    from comparing phrases.
    """

    def cast(self, data):
        # If all entries have the same number of entries, I treat them as a
        # table. Then I can use column-based type inference.
        if len({len(xs) for xs in data}) == 1:
            N = len(data[0])
            log(f"Applying column type inference (all headers have {N-1} fields)")
            tabular_data = [[row[i] for row in data] for i in range(N)]
            return headlessTabularTyping(
                tabular_data, levels=self.levels, na_str=self.na_str
            )
        else:
            return [
                Phrase(
                    [Datum(x, na_str=self.na_str).data for x in row], levels=self.levels
                )
                for row in data
            ]

    def parse(self, filehandle):
        """
        Return a list of lists of strings. Currently only FASTA is supported.
        """
        return self._parse_fasta(filehandle, sep="|")

    def _parse_fasta(self, filehandle, sep="|"):
        """
        Parse a fasta file. The header is split into fields on 'sep'. The
        sequence is added as a final field.
        """
        p_header = parsec.string(">") >> parsec.regex("[^\n\r]*") << parsec.spaces()
        p_seq = (
            parsec.sepBy1(
                parsec.regex("[^>\n\r]*"), sep=parsec.regex("[\r\n\t ]+")
            ).parsecmap(concat)
            << parsec.spaces()
        )
        p_entry = p_header + p_seq
        p_fasta = parsec.many1(p_entry)
        log(f"Reading {file_str(filehandle)} as a fasta file:")
        try:
            entries = p_fasta.parse(filehandle.read())
        except AttributeError:
            # in case I want to pass in a list of strings, e.g., in tests
            entries = p_fasta.parse(filehandle)
        row = [h.split(sep) + [q] for (h, q) in entries]
        return row


#  class HetList(Interpreter):
#      """
#      Interpret a list of items of different types
#      """
#      def cast(self, items):
#          pass


#  class Nested(Interpreter):
#      """
#      Interpret a nested data structure (e.g. JSON)
#      """
#      def cast(self, nest):
#          pass


class Phrase:
    def __init__(self, tokens, levels=None):
        self.tokens = tokens
        self.levels = levels

    def connect(self, g, taguri=None):
        """
        Create links between a list of Tokens. For example, they may be related
        by fields in a fasta header or elements in a row in a table.
        """
        for token in self.tokens:
            if token.clean is None:
                continue
            if (self.levels is None) or (token.group in self.levels):
                token.relate(self.tokens, g, levels=self.levels)
            token.add_triples(g)
            if taguri and token.group:
                turi = token.as_uri()
                if turi:
                    g.add((turi, P.tag, taguri))

    def __str__(self):
        return str([(t.typename, t.field_name, t.clean) for t in self.tokens])



from octofludb.util import compose, underscore, lower, strip


def make_flat_wordfinder(
    wordlist,
    alphabet="abcdefghijklmnopqrstuvwxyz",
    depth=1,
    clean=compose(underscore, lower, strip),
):
    """
    Build a function for finding the closest word in a list. This is not a
    spellchecker, since it will return None if not word is found. Also, it is
    not limited to words, but includes phrases. I may adapt it in the future to
    allow the deletion of words.
    """

    WORDS0 = {clean(w) for w in wordlist}
    WORDSN = [{e: w for w in WORDS0 for e in edits(w, alphabet)}]
    for i in range(2, depth + 1):
        WORDSN.append(
            {e2: w for (e1, w) in WORDSN[-1].items() for e2 in edits(e1, alphabet)}
        )

    def wordfinder(word):
        clean_word = clean(word)

        if clean_word in WORDS0:
            return clean_word

        for words in WORDSN:
            if clean_word in words:
                return words[clean_word]

        return None

    return wordfinder


def edits(word, alphabet):
    """
    Borrowed directly from Peter Norvig's spell checker (https://norvig.com/spell-correct.html).
    """
    splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [L + R[1:] for L, R in splits if R]
    transposes = [L + R[1] + R[0] + R[2:] for L, R in splits if len(R) > 1]
    replaces = [L + c + R[1:] for L, R in splits if R for c in alphabet]
    inserts = [L + c + R for L, R in splits for c in alphabet]
    return set(deletes + transposes + replaces + inserts)



import sys
import time
import os
import requests
import datetime
from typing import List
from urllib.error import HTTPError
from Bio import Entrez
from tqdm import tqdm
from octofludb.util import log
import octofludb.colors as colors
import pgraphdb as db

Entrez.email = "zebulun.arendsee@usda.gov"


def get_all_acc_in_db(url="http://localhost:7200", repo="octofludb"):

    sparql_filename = os.path.join(os.path.dirname(__file__), "data", "all-acc.rq")

    acc = db.sparql_query(sparql_file=sparql_filename, url=url, repo_name=repo)

    return [x["acc"]["value"] for x in acc["results"]["bindings"]]


def get_acc_by_date(
    mindate, maxdate, ignore=[], retmax=100000, query='"Influenza+A+Virus"[Organism]'
):
    """
    mindate: a date string of form YYYY/MM, e.g. "2020/01"
    maxdate: a date string of form YYYY/MM, e.g. "2020/06"
    """
    base = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
    params = {
        "db": "nuccore",
        "term": query,
        "retmode": "json",
        "retmax": str(retmax),
        "datetype": "pdat",
        "mindate": mindate,
        "maxdate": maxdate,
        "idtype": "acc",
    }

    req = requests.get(base, params=params)
    try:
        result = req.json()["esearchresult"]

        if int(result["retmax"]) < int(result["count"]):
            log(
                f'{colors.bad("Warning:")} results truncated at {result["retmax"]} of {result["count"]} ids'
            )
    except:
        log(f'{colors.bad("Error:")} could not find "esearchresult"')
        log(str(req))
        log(params)
        return []

    # For great manner
    time.sleep(1)

    return result["idlist"]


def missing_acc_by_date(
    min_year=1918,
    max_year=2099,
    nmonths=9999,
    url="http://localhost:7200",
    repo="octofludb",
):
    """
    Find all genbank accessions that are missing from the database. Return as a tuples of (date, [accession])
    """
    now = datetime.datetime.now()
    cur_year, cur_month = now.year, now.month

    old_acc = {s for s in get_all_acc_in_db(url=url, repo=repo)}

    # step backwards in time, month-by-month to year 2000
    for year in reversed(range(2000, cur_year + 1)):
        if year < min_year:
            break
        if year > max_year:
            continue
        for month in reversed(range(1, 12 + 1)):
            if nmonths <= 0:
                break

            if year == cur_year and month > cur_month:
                # pulling future sequences is not yet supported
                continue
            mth_acc = get_acc_by_date(
                mindate=f"{str(year)}/{str(month)}", maxdate=f"{str(year)}/{str(month)}"
            )
            new_acc = [acc for acc in mth_acc if not acc in old_acc]
            nmonths -= 1
            yield (f"{str(year)}/{str(month)}", new_acc)

    # step backwards in time, year-by-year to year 1918
    for year in reversed(range(1918, 2000)):
        if year < min_year or nmonths <= 0:
            break
        if year > max_year:
            continue
        year_acc = get_acc_by_date(mindate=str(year), maxdate=str(year))
        new_acc = [acc for acc in year_acc if not acc in old_acc]
        #  nmonths -= 12
        yield (str(year), new_acc)


# code adapted from http://biopython.org/DIST/docs/tutorial/Tutorial.html#htoc122
def get_gbs(gb_ids: List[str]) -> List[dict]:
    batch_size = 1000
    count = len(gb_ids)
    for start in tqdm(range(0, count, batch_size)):
        end = min(count, start + batch_size)
        attempt = 0
        while attempt < 10:
            try:
                h = Entrez.efetch(db="nucleotide", id=gb_ids[start:end], retmode="xml")
                x = Entrez.read(h)
                h.close()
                yield x
                break
            except Exception as err:
                attempt += 1
                print(f"Received error from server {err}", file=sys.stderr)
                print(f"Attempt {str(attempt)} of 10 attempt", file=sys.stderr)
                time.sleep(15)



import argparse
import textwrap


class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """
    Remove the redundant "<subcommand>" string from under the "subcommands:"
    line in the help statement.

    Adapted from Jeppe Ledet-Pedersen on StackOverflow.
    """

    def _format_action(self, action):
        parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


# subcommand decorator idea adapted from Mike Depalatis blog
def subcommand_maker(subparsers):
    def subcommand(args=[], parent=subparsers):
        def decorator(func):
            if func.__doc__:
                help_str = func.__doc__.strip().split("\n")[0]
                desc_str = textwrap.dedent(func.__doc__)
            else:
                help_str = "DOCUMENT ME PLEASE!!!"
                desc_str = None
            cmd_name = args[0]
            parser = parent.add_parser(
                cmd_name,
                description=desc_str,
                help=help_str,
                formatter_class=argparse.RawDescriptionHelpFormatter,
            )
            for arg in args[1:]:
                parser.add_argument(*arg[0], **arg[1])
            parser.set_defaults(func=func)

        return decorator

    return subcommand


def argument(*name_or_flags, **kwargs):
    return (list(name_or_flags), kwargs)
